////////////////////////////////////////////////////////////////////////////////
// cb1.v  -  c933hdmi project
////////////////////////////////////////////////////////////////////////////////
// cb1.v
// Top level module for the project.
// Contains code for:
// Instantiating sub modules.
// Reading and writing SRAM, BRAM, distributed RAM and arrays of registers.
// SPI controller.
// Input/output interface for flashing the M25P80 memory chip.
// Synchronizing Apple signals to CB clock
// Slot device I/O ($C0nx)
// DMA & interrupt priority daisy chain 
////////////////////////////////////////////////////////////////////////////////
// c933hdmi project
// This version is for hdmi using the hdmi/dvi peripheral board for the 
// Carte Blanche card.  It is for an Apple IIgs only.  
// The 12-bit mod is not supported.  6-bit will still be available 
// The resolution is 720x480.  The DCM (main clock) output must be ~56.644 MHz 
// Signals or clocks that are synchronized with the main clock appended with _s
// Fastflash code is included but undefined for now because of higher clock
// M25 flash RAM code is included but is not usable with higher clock
// charROM distributed ROM replaces the font block RAM
////////////////////////////////////////////////////////////////////////////////
`include "c933hdmi_defs.vh" // contains global definitions
////////////////////////////////////////////////////////////////////////////////
module C933hdmi             // For the Apple IIgs only
  (     
////////////////////////////////////////////////////////////////////////////////
// inputs/outputs to module
// labels starting with BUS_ are Apple II slot bus signals                      
   input    BUS_3M58,       // M2B0 on a IIgs (slot 3 on ROM 01, slots 1-6 on ROM 03, CREF on slot 7)
   input        BUS_7M,         // 7 MHz Apple clock
   inout [15:0]  BUS_A,     // Apple address bus
   inout [7:0]  BUS_D,      // Apple data bus
    output  BUS_DMA,            // Direct Memory Access
   input    BUS_DMAIN,      // DMA priority daisy chain
   output   BUS_DMAOUT,     // DMA priority daisy chain
    output   BUS_INH,           // inhibit
   input    BUS_INTIN,      // interrupt priority daisy chain
   output   BUS_INTOUT,     // interrupt priority daisy chain
   input    BUS_IOSTROBE,   // input/output strobe 
   input    BUS_IRQ,            // interrupt request
   input    BUS_NMI,            // non-maskable interrupt
   input    BUS_PHASE0,     // phase 0 clock
   input    BUS_PHASE1,     // phase 1 clock
   input    BUS_Q3,         // Q3 clock
   input    BUS_R_W,            // read/write signal
   output   BUS_RDY,            // ready
   input    BUS_RESET,      // reset
   input    BUS_SYNC,       // composite h & v sync (slot 7 only)
   input    BUS_USER1,      // /M2SEL on a IIgs
   input    N_BUS_DEV_SEL,  // Apple slot device select
   input    N_BUS_IO_SEL,   // Apple slot io select 
// --- hdmi/dvi PB --- //
    output [7:0] RED,           // 8 bit red to hdmi    
    output [7:0] GREEN,     // 8 bit green to hdmi  
    output [7:0] BLUE,      // 8 bit blue to hdmi   
    output HSYNC,               // horizontal sync pulse to hdmi
    output VSYNC,               // vertical sync pulse to hdmi
    output FRAME_ENABLE,        // frame enable signal to hdmi
    output PIX_CLK,         // hdmi pixel clock
    output COLOR_LEVEL,     // color level (1 = 24 bit, 0 = 12 bit)
// ------------------- //
   output [16:0]   SRAM_A,  // SRAM address
   inout  [7:0]    SRAM_D,  // SRAM data
   output   SRAM_NCS0,      // SRAM enable chip 0 (low)     
   output   SRAM_NCS1,      // SRAM enable chip 1 (low)
   output   SRAM_NOE,       // SRAM output enable (low)
   output   SRAM_NWE,       // SRAM write enable    (low)   
   // input    SYSCLK,         // 14.31818 MHz onboard CB clock

   input   clk50,
   input   clk25,
   input LOCKED_OUT,
   // buffers WE
   output video_we,   // enable writing
   // charROM
   output [10:0] rom_addr,
   input [7:0] char_data,
   // PAL BUF
   output [7:0] BRAM_AA_PAL,    // write address 8 bit
   output [11:0] BRAM_DA_PAL,         // write data 12 bit
   output [7:0] BRAM_AB_PAL,    // read address 8 bit
   input [11:0] BRAM_DB_PAL,   // read data 12 bit
   //text buf
    output [11:0] BRAM_AA,    // write address 12 bit
    output [7:0] BRAM_DA,     // write data
    output [11:0] BRAM_AB,    // read address 12 bit
    input [7:0] BRAM_DB, // read data
    // HGR BUF
    output [13:0] BRAM_AA_hgr,   // write address 14 bit
    output [7:0] BRAM_DA_hgr,    // write data
    output [13:0] BRAM_AB_hgr,   // read address 14 bit
    input [7:0] BRAM_DB_hgr,  // read data


   output [5:0]  RGB,       // red-green-blue for 6 bit VGA color
   output   v_sync,         // horizontal sync pulse for 6 bit VGA color
   output   h_sync          // vertical sync pulse for 6 bit VGA color
`ifdef m25
    ,
   input    spi_sdo,        // SPI - serial data out to device
   output   spi_sdi,        // SPI - serial data into device
   output   spi_sck,        // SPI - clock for device
   output   reg spi_rom_cs  // SPI - chip select
`endif // m25
  );
////////////////////////////////////////////////////////////////////////////////
// version 
    reg [7:0] maj_version = 9;  // use 255 for test versions
    reg [7:0] min_version = 33;
////////////////////////////////////////////////////////////////////////////////
// registers used to sync Apple clocks to clk50 
    reg reset_s = 1;     
    reg phase0_s = 0;   
    reg Q3_s = 1;       
    reg devsel_s = 1;   
    reg m2b0_s = 0;     
    reg[15:0] bus_a_1 = 0;  
    reg[15:0] bus_a_s = 0;  
    reg[7:0] bus_d_1 = 0;       
    reg[7:0] bus_d_s = 0;
    reg user1_s =0;
    reg ph1_s = 1;  
    reg rw_s = 0;
    reg bus_7M_s = 1;
    reg iosel_s = 1;
    reg rw_1 = 0;
    reg reset_1 = 1;
    reg phase0_1 = 0;
    reg Q3_1 = 1;
    reg devsel_1 = 1;
    reg m2b0_1 = 0;
    reg ph1_1 = 1;
    reg user1_1 = 0;
    reg bus_7M_1 = 1;
    reg iosel_1 = 1;
////////////////////////////////////////////////////////////////////////////////
`ifdef bbug
    wire bugin;                         // used for debugging
`endif 
`ifdef fastflash
    reg[8:0] ff_addr = 9'h000;      // fast flash uses hgr2 addr
    reg[8:0] fr_addr = 9'h000;      // read back & error check
`endif
////////////////////////////////////////////////////////////////////////////////
// video buffer read/write timing
   wire     video_read;                 // allow video SRAM read (from vga.video.v)
  //SH wire     video_we;                   // write enable SRAM (from vga.video.v)
   wire     video_write_SRAM;           // time to write data from Apple
////////////////////////////////////////////////////////////////////////////////
// sync
    wire h_sync2;                       // h sync from vga.video.v
    wire v_sync2;                       // v sync from vga.video.v
    assign HSYNC = h_sync2;         // h sync for hdmi
    assign VSYNC = v_sync2;         // v sync for hdmi
    assign h_sync = h_sync2;        // h sync for 6 bit VGA
    assign v_sync = v_sync2;        // v sync for 6 bit VGA
////////////////////////////////////////////////////////////////////////////////
// DCM (clk0.xaw) input/output wires
// remove sh   wire     CLKFX_OUT;                  // DCM output frequency
  //  SH wire     LOCKED_OUT;                 // high when DCM has achieved lock (proper output)
////////////////////////////////////////////////////////////////////////////////
// instantiate main CB clock (DCM)
// If you need to use the input clock for something else put the following
// line in the list below:
// .CLKIN_IBUFG_OUT(CLKIN_IBUFG_OUT),
// and define a wire .CLKIN_IBUFG_OUT

//   clk0 clock (
//             .CLKIN_IN(SYSCLK),             // input from onboard CB 14.3MHz source
//           .RST_IN(1'b0),               // causes DCM to be reset only on startup
//           .CLKFX_OUT(CLKFX_OUT),   // output frequency ~57.2727MHz
//           .LOCKED_OUT(LOCKED_OUT)  // high when DCM achieves the proper clock
//           );
////////////////////////////////////////////////////////////////////////////////
// clocks
 // SH  wire clk50;          // main clock
 // SH   wire clk25;                         // clk50 divided by 2
////////////////////////////////////////////////////////////////////////////////
`ifdef tester
    wire [9:0] test00;              // used for keyboard input to do internal tests
`endif
////////////////////////////////////////////////////////////////////////////////
// INH not used in this version
// Apple IIgs must use mask when asserting INH - see IIgs technote #32                  
    assign  BUS_INH = 1'bz;             // not used set to high-z
////////////////////////////////////////////////////////////////////////////////
// priority daisy chains - connect pins
   assign   BUS_DMAOUT = BUS_DMAIN;     // just connect the DMA chain 
    assign  BUS_INTOUT = BUS_INTIN;     // just connect the interrupt chain
////////////////////////////////////////////////////////////////////////////////
// DMA not used
    assign  BUS_DMA = 1'bz;             // not used set to high-z   
////////////////////////////////////////////////////////////////////////////////    
// RDY not used
    assign  BUS_RDY = 1'bz;             // not used set to high-z
////////////////////////////////////////////////////////////////////////////////    
// RGB from module vga_video.v and sent out to 6 bit VGA circuit
   wire [5:0]  RGB_from_module;                      
    assign RGB = RGB_from_module;
////////////////////////////////////////////////////////////////////////////////    
// reset for CB code - if Apple resets or DCM is not locked (not proper output)
    wire reset = ~reset_s | ~LOCKED_OUT;    // this reset wire is normally low
////////////////////////////////////////////////////////////////////////////////    
// video buffer access wires
// sh   wire [10:0] rom_addr;            // address of character data from vga_video.v
// sh   wire [7:0]   char_data;          // character sent to vga_video.v
   wire [7:0]   video_ram_data; // data read from CB video buffers
   wire [16:0] ram_address;     // address used to read from video buffer  
   wire     auxmem;                         // 1 means the apple wrote a bank 1 video
                                            // address (aux memory) - calculate in vga_video.v
////////////////////////////////////////////////////////////////////////////////    
// write clock - from old wrclk module
   wire     update_time;                // OK to update the CB buffer   (from vga_video.v)
   wire [15:0] video_wr_addr;   // address of data in Apple and CB buffer
   wire [7:0]   write_data;         // data in Apple and written to CB buffer
   reg [15:0]   pending_addr = 0;   // hold bus address until ready
   reg [7:0]    pending_data = 0;   // hold data till ready
   reg [3:0]    wc_state = 4'b0000;         // state machine state  
   reg [3:0]    next_wc_state = 4'b0000;    // state machine next state
    reg video_we_reg = 0;           // write enable SRAM        
   reg write_from_6502;             // used in state machine
////////////////////////////////////////////////////////////////////////////////    
// write clock time 
// on negative edge of Q3 and phase0 is high and user1 is low (7M was added to improve timing 1/2015)
    wire wr_cl = (Q3_s & ~Q3_1) && phase0_s && (user1_s == 0) && bus_7M_s; 
////////////////////////////////////////////////////////////////////////////////    
// get the address and data from the Apple bus    
   always @ (posedge clk50) 
        begin
      write_from_6502 <= ((wr_cl == 1) & (rw_s == 0));      // was blocking 1/1/15 
      wc_state <= next_wc_state;            // update the state
      if(write_from_6502) 
            begin
            pending_addr <= bus_a_s;        // get the address
            pending_data <= bus_d_s;        // get the data
            end     
        end
////////////////////////////////////////////////////////////////////////////////    
// write enable SRAM buffer
    always @ (negedge clk50) video_we_reg <= (wc_state == 4);   
////////////////////////////////////////////////////////////////////////////////    
// state machine to determine when its time to write to SRAM
   always @(*)  
        begin
      case (wc_state)
            4'h0:
                if(write_from_6502)     // Apple writing to bus?
                next_wc_state = 1;      // yes
                else
                next_wc_state = 0;      // loop back here till it does
            4'h1:
                if(update_time)             // too late, wait for the next slot
                next_wc_state = 1;      // loop here until not update time
                else
                next_wc_state = 2;      // not update time 
            4'h2:
                if(update_time)         // update time?
                next_wc_state = 3;      // yes
                else
                next_wc_state = 2;      // no wait for update time
            4'h3: next_wc_state = 4;    // add some delay
            4'h4: next_wc_state = 5;
            4'h5:   
                next_wc_state = 0;
//              next_wc_state = 6;   // adding states 6 - 9 helps remove spurious persistant pixels in SHR VGA

            4'h6: next_wc_state = 7;
            4'h7: next_wc_state = 8;
            4'h8: next_wc_state = 0;
            default:
                next_wc_state = 0;
      endcase 
        end 
////////////////////////////////////////////////////////////////////////////////    
    assign video_we = video_we_reg;
    assign video_write_SRAM = (wc_state == 3) || (wc_state == 4) || (wc_state == 5) ; 
   assign write_data = pending_data;            // Apple video data
    assign video_wr_addr = pending_addr;        // Apple RAM address of video data

////////////////////////////////////////////////////////////////////////////////
// Main memory video data from $2000 to $5FFF is buffered in FPGA
// BlockRAM (BRAM) - hgr_buf
//
// write_vid_is_hgr 
// high if Apple is writing to video address $2000 to $5FFFF in main memory, not aux memory
    wire write_vid_is_hgr = (video_wr_addr > 16'h1FFF) & (video_wr_addr < 16'h6000) & ~auxmem;
    
// read_vid_is_hgr
// high if buffer read address (from vga_video.v) is $2000 to $5FFF in main memory, not aux memory
    wire read_vid_is_hgr = (ram_address[15:0] > 16'h1FFF) & (ram_address[15:0] < 16'h6000) & (ram_address[16] == 1'b0);
    
// Block RAM buffer I/O for main memory video data $2000 to $5FFFF  
// sh    wire[13:0] BRAM_AA_hgr; // write address
// sh    wire[13:0] BRAM_AB_hgr; // read address
// sh    wire[7:0] BRAM_DA_hgr;      // write data
// sh    wire[7:0] BRAM_DB_hgr;      // read data

// instantiate - hgr_buf - simple dual port RAM
 //   hgr_buf hgr_buf(                // buffers main RAM HGR & DHR both pages
 //       .clka(clk50),               // clock for writing
 //       .wea(video_we),             // enable writing
 //       .addra(BRAM_AA_hgr),    // write address 14 bit
 //       .dina(BRAM_DA_hgr),         // write data
 //       .clkb(clk50),               // clock for reading
  //      .addrb(BRAM_AB_hgr),    // read address 14 bit
   //     .doutb(BRAM_DB_hgr));   // read data 

////////////////////////////////////////////////////////////////////////////////
// Text memory data from $400 to $0BFF main & aux is buffered in FPGA 
// BlockRAM (BRAM) - text_buf
// 
// write_vid_is_text
// high if Apple is writing address $0400 to $0BFF (TEXT pg 1 or 2 in either bank 
    wire write_vid_is_text = (video_wr_addr > 16'h03FF) & (video_wr_addr < 16'h0C00);

// read_vid_is_text
// high if buffer read address (from vga_video.v) is $0400 to $0BFF in either bank 0 or 1
    wire read_vid_is_text = (ram_address[15:0] > 16'h03FF) & (ram_address[15:0] < 16'h0C00);
    
// Block RAM buffer I/O for text memory 
 //   wire[11:0] BRAM_AA; // write addr
 //   wire[11:0] BRAM_AB; // read addr
 //   wire[7:0] BRAM_DA;  // write data
 //   wire[7:0] BRAM_DB;  // read data

// instantiate - text_buf - simple dual port RAM
  //  text_buf text_buf ( // buffers text pages 1 & 2, main and aux
   //     .clka(clk50),       // clock for writing
   //     .wea(video_we),   // enable writing
   //     .addra(BRAM_AA),    // write address 12 bit
   //     .dina(BRAM_DA),     // write data
   //     .clkb(clk50),       // clock for reading
   //     .addrb(BRAM_AB),    // read address 12 bit
    //    .doutb(BRAM_DB)); // read data
    
////////////////////////////////////////////////////////////////////////////////    
// Palettes (in SHR) $019E00 to $019FFF are buffered in FPGA
// BlockRAM (BRAM) - pal_buf
//
// write_is_pal
// high if Apple is writing address $019E00 to $019FFF
    wire write_is_pal =  ((video_wr_addr > 16'h9DFF) && (video_wr_addr < 16'hA000) && auxmem ) /* ? 1'b1 : 1'b0*/;
// write_is_even 
// high if the byte written is even
    wire write_is_even = (video_wr_addr[0] == 1'b0) /*? 1'b1 : 1'b0*/;
    
// Block RAM buffer I/O for SHR palettes
//    wire[7:0] BRAM_AA_PAL;  // write addr
//    wire[7:0] BRAM_AB_PAL;  // read addr
//    wire[11:0] BRAM_DA_PAL; // write data
//    wire[11:0] BRAM_DB_PAL; // read data
    
// instantiate - pal_buf - simple dual port RAM
// sh    pal_buf pal_buf (               // buffers palettes 12 bits
// sh        .clka(clk50),               // clock for writing
// sh        .wea(video_we),             // enable writing
// sh        .addra(BRAM_AA_PAL),    // write address 8 bit
// sh        .dina(BRAM_DA_PAL),         // write data 12 bit
// sh        .clkb(clk50),               // clock for reading
 // sh       .addrb(BRAM_AB_PAL),    // read address 8 bit
 // sh       .doutb(BRAM_DB_PAL));   // read data 12 bit

// use palette address to get palette data
    wire[7:0] pal_addr; // read address from vga_video.v to BRAM - pal_buf
    wire[11:0] pal_data; // read data to vga_video.v from BRAM - pal_buf    
    
// odd-even palette registers
// palettes are 12-bits store in 2 bytes (16 bits) 4 bits in odd bytes and
// 8 bits in even bytes
    reg [7:0] even_pal = 0; // 8 bits       
    reg [3:0] odd_pal = 0;  // 4 bits

// palette address written to BRAM - pal_buf
    assign BRAM_AA_PAL = (video_write_SRAM && write_is_pal) ? (video_wr_addr[8:1]) : 8'bz;
// palette data written to BRAM - pal_buf
    assign BRAM_DA_PAL = (video_write_SRAM && write_is_pal) ? {odd_pal, even_pal} : 12'bz;
// palette address for reading palette data from BRAM - pal_buf
    assign BRAM_AB_PAL = pal_addr;
// palette data read from BRAM - pal_buf
    assign pal_data = BRAM_DB_PAL;
////////////////////////////////////////////////////////////////////////////////    
`ifdef fastflash
    wire write_vid_is_ff = ((video_wr_addr > 16'h3FFF) && (video_wr_addr < 16'h4100) && ~auxmem && video_write_SRAM);
`endif
////////////////////////////////////////////////////////////////////////////////
// SRAM chip select 
   assign   SRAM_NCS0 = 0;                  // use SRAM chip 0
   assign   SRAM_NCS1 = 1;                  // do not use SRAM chip 1
////////////////////////////////////////////////////////////////////////////////
// SRAM - set write enable and output enable 

    assign  SRAM_NOE =  video_we /*& ~video_read */;    // output enable
   assign   SRAM_NWE = ~video_we;                           // write enable
////////////////////////////////////////////////////////////////////////////////
// SRAM reading & writing
    reg [16:0] SRAM_A_reg = 0;      // hold read/write address 
    reg [7:0] SRAM_D_reg = 0;       // hold read/write data
// video_mem
// high when HGR, DHR or SHR data is written by Apple
    wire video_mem = ((video_wr_addr > 16'h1FFF) && (video_wr_addr < 16'h9D00)) /*? 1'b1 : 1'b0*/;
// vid_read_mem
// high when HGR, DHR or SHR data is being read by VGA/HDMI (vga_video.v) 
    wire vid_read_mem = ((ram_address[15:0] > 16'h1FFF) && (ram_address[15:0] < 16'h9D00)) /*? 1'b1 : 1'b0 */;
    
// write_vid_is_gs
// high when SHR palettes or scan control bytes are being written by Apple
    wire write_vid_is_gs = ((video_wr_addr > 16'h9CFF) && (video_wr_addr < 16'hA000) && auxmem) /*? 1'b1 : 1'b0*/;

// write to SRAM
    always @(negedge clk50)
        begin
        if (clk25)
            begin
            SRAM_A_reg <= (video_write_SRAM && ~write_vid_is_text && ~write_vid_is_hgr  && ~write_vid_is_gs) ?
                                {auxmem, video_wr_addr} : (vid_read_mem  && ~read_vid_is_hgr) ? ram_address : 17'bz;
            SRAM_D_reg <= (video_write_SRAM & video_mem & ~write_vid_is_hgr && ~write_vid_is_gs) ? write_data : 8'bz;   
            end
        end

    assign  SRAM_A = SRAM_A_reg;
    assign  SRAM_D = SRAM_D_reg;
////////////////////////////////////////////////////////////////////////////////    
// this stuff needs work 
// I don't think tri-states are allowed internal to FPGA!
// Maybe clock this stuff with main clock?

    assign  video_ram_data = (~video_write_SRAM & vid_read_mem & ~read_vid_is_hgr) ?  SRAM_D : read_vid_is_text ? BRAM_DB : (~video_write_SRAM & read_vid_is_hgr) ? BRAM_DB_hgr : 8'bz; 

    assign  BRAM_AA = (video_write_SRAM && write_vid_is_text) ? {auxmem, video_wr_addr[10:0]} : 12'bz;  // write addr
    assign  BRAM_AB = (read_vid_is_text) ? {ram_address[16], ram_address[10:0]} : 12'bz;                        // read addr
    assign  BRAM_DA = (video_write_SRAM && write_vid_is_text) ? write_data : 8'bz;

    assign  BRAM_AA_hgr = (video_write_SRAM && write_vid_is_hgr) ? video_wr_addr[13:0] : 14'bz;     // write
    assign  BRAM_AB_hgr = (read_vid_is_hgr) ? ram_address[13:0] : 14'bz;                        // read addr    
    assign  BRAM_DA_hgr = (video_write_SRAM && write_vid_is_hgr) ? write_data : 8'bz;
////////////////////////////////////////////////////////////////////////////////    
// scan control bytes (scb) are handled in an array of registers    
    wire[7:0] scb_addr; // read address
    wire[7:0] scb_data; // read data
    reg [7:0] scb [199:0];  // array of 200, 8 bit regs

// create the array of registers
    integer k;
    initial
        begin
        for (k = 0; k <= 199; k = k + 1) scb[k] = 0;
`ifdef fastflash
        for (k = 0; k <= 255; k = k + 1) ff_array [k] = 0;
`endif          
        end

    assign scb_data = scb[scb_addr];

// write_is_scb
// high if Apple is writing to address $019D00 to $019DC7 (200 bytes)   
    wire write_is_scb = ((video_wr_addr > 16'h9CFF) && (video_wr_addr < 16'h9DC8) && auxmem) /*? 1'b1 : 1'b0 */;
    
`ifdef fastflash
reg [7:0] ff_array [255:0];
`endif

    always @(posedge clk50)
        begin
        if (video_write_SRAM && write_is_scb) scb [video_wr_addr[7:0]] <= write_data;       // data is scb
        if(video_write_SRAM && write_is_pal && write_is_even) even_pal <= write_data;           // data is even palette
        if(video_write_SRAM && write_is_pal && ~write_is_even) odd_pal <= write_data[3:0];  // data is odd palette
`ifdef fastflash
        if(write_vid_is_ff) ff_array[video_wr_addr[7:0]] <= write_data;                         // data is for fastflash 
`endif
        end
////////////////////////////////////////////////////////////////////////////////
// instantiate the vga_video.v module
   vga_video 
     video_module (
           .clk50(clk50),                   // ~57.2727MHz from DCM (higher than 56.644 MHz it should be)
           .char_data(char_data),           // from charROM (distributed ROM is used instead of block RAM)
           .ram_data(video_ram_data),       // from video data SRAM
           .bus_addr(bus_a_s),              // Apple slot bus address - sync'd
            .bus_data(bus_d_s),             // Apple slot bus data - sync'd
           .ph1_s(ph1_s),                       // Apple phase1 clock - sync'd
            .phase0_s(phase0_s),            // Apple phase0 clcok - sync'd
            .Q3(Q3_s),                          // Apple Q3 clock - sync'd in cb1.v
//         .rw(BUS_R_W),                        // read/write from Apple
            .rw_s(rw_s),                        // read/write from Apple - sync'd in cb1.v
           .reset_n(~reset),                // Apple reset inverted (high when reset)
            .user1_s(user1_s),              // on a IIgs low means first 128K of RAM - sync'd in cb1.v
            .devsel_s(devsel_s),                // Apple slot /DEVSEL           
            .m2b0_s(m2b0_s),                    // this signals bank 0/1
`ifdef bbug         
            .bugin(bugin),                      // added to pass signal for bug border
`endif
            .scb_data(scb_data),                // input scanline from BRAM
            .scb_addr(scb_addr),                // output scanline address in BRAM
            .pal_data(pal_data),                // input palette even byte from BRAM
            .pal_addr(pal_addr),                // output palette even byte address in BRAM
           .update_time(update_time),       // when it's OK to write Apple data to SRAM
            .auxmem(auxmem),                    // high when mirroring auxram (bank1) to CB RAM 
           .video_read(video_read),     // high when it's OK to read the CB RAM
           .v_sync(v_sync2),                // for 6 bit & hdmi
           .h_sync(h_sync2),                // for 6 bit & hdmi
            .RED(RED),                          // 8 bits for 24 bit hdmi
            .GREEN(GREEN),                      // 8 bits for 24 bit hdmi
            .BLUE(BLUE),                        // 8 bits for 24 bit hdmi
            .FRAME_ENABLE(FRAME_ENABLE),    // high when pixels are drawn
            .PIX_CLK(PIX_CLK),              // 28.322 MHz is std. for 720x480
            .COLOR_LEVEL(COLOR_LEVEL),      // 1 = 24 bit, 0 = 12 bit hdmi
           .rom_addr(rom_addr),             // address used to get character data
           .ram_address(ram_address),       // address to get ram_data from CB RAM buffer
`ifdef tester
            .wtest00(test00),                   // 0 - 1023 register used for debug testing
`endif
        // sh   .clk25(clk25),                      // half of clk50 - for 720x480 should be ~28.322 MHz
            .RGB(RGB_from_module)           // 6 bit RGB used for the original CB 6-bit VGA connector
           );

////////////////////////////////////////////////////////////////////////////////
// instantiate distributed ROM for text characters           
   // charROM charROM(
   //        .a(rom_addr),        // address from vga_video.v
   //        .spo(char_data), // character data for text in vga_video.v
   //        .clk(clk50)      // main clock
 //          );
`ifdef m25
////////////////////////////////////////////////////////////////////////////////
// SPI
    reg clk25_b;                // clock for spi code & spi controller
    wire sck;                   // spi clock (clk25_b) that can be enabled/disabled
    reg mosi = 0;               // spi serial data input
    assign spi_sdi = mosi;  // to spi_sdi (serial data into device)
   assign spi_sck = sck;    // to spi_sck (clock for device)
                                    // device is NVRAM (flash) chip
    wire [7:0] spi_out;
    wire [7:0] spi_in;
    wire write_spi;
    wire spi_done;
// The following spi code is taken from the spi.v module.
// See copyright and license below:
// ====================================================================
//                         VECTOR-06C FPGA REPLICA
//
//      Copyright (C) 2007, Viacheslav Slavinsky
//
// This core is distributed under modified BSD license. 
// For complete licensing information see LICENSE.TXT.
// --------------------------------------------------------------------  
   wire miso = spi_sdo;     // from spi_sdo (serial data from device)
   wire [7:0] di = spi_in;      // 8 bit parallel data in
   wire wr = write_spi;     // 1 = write to spi
   wire [7:0] do;                   // 8 bit parallel data out
    assign spi_out = do;       // 8 bit wire do to spi_out
   reg  done;                       // 1 = 8 bits shifted in/out
    assign spi_done = done;     // spi routine is done
   reg [7:0] shiftreg;
   reg [7:0] shiftski;
   reg [1:0] state = 0;
   reg scken = 0;

   assign sck = clk25_b & scken;                        // scken used to switch on/off device clock
   assign do  = shiftreg;

   always @(posedge clk50 or negedge reset_s)   // this block functions on every rising edge
                                                                // of the 50Mz clock or on the falling edge
                                                                // of the Apple RESET the '_s' means sync'd 
                                                                // with 50 Mz clock
        begin
      if (!reset_s)             // if reset then set the registers to zero
            begin
            state <= 0;         // reset state
            mosi <= 1'b0;       
            done  <= 0;         // not done
            scken <= 0;         // device clock off
            end 
        else                        // every positive edge of 50Mz clock
            begin
            if (clk25_b)        // every other time (25Mz)
                begin
                case (state)    // state machine to do the actual work
                0:  begin           // state 0 we start here
                    if (wr)         // if 1 then write otherwise do nothing
                        begin       
                        done <= 1'b0;                   // means not done           
                        state <= 1;                     // next time through we start with state 1
                        shiftreg <= di;             // load the data we are going to write
                        shiftski <= 8'b11111111;    // use as counter (initialize it)
                        end
                    end
                1: begin                                    // state 1
                    scken <= 1;                         // device clock on (the device we writing to)
                    mosi <= shiftreg[7];                // send a bit 7 to device
                    shiftreg <= {shiftreg[6:0],miso};   // left shift data & shift bit in from device
                                                                    // notice shiftreg holds writes & reads at 
                                                                    // the same time
                    shiftski <= {1'b0,shiftski[7:1]};   // right shift 1 bit out of counter
                    if (|shiftski == 0)             // tests for 0 in counter, if not 0 then we loop
                                                            // state 1 on next 25Mz cycle
                        begin                           // counter is 0
                        state <= 2;                     // next cycle we go directly to state 2
                        scken <= 0;                     // device clock off (device can no longer 
                                                            // accept data input
                        end
                    end
                2:  begin               // state 2
        //          mosi <= 1'b0; // shouldn't be necessary but a nice debug view (VS)
                    done <= 1'b1; // tell everybody we're done with this byte 
                    state <= 0;   // next byte we start over with state 0
                    end
                default: ;       // you should always have a default in a case statement
                endcase
            end
      end
   end
`endif // m25
////////////////////////////////////////////////////////////////////////////////
// bus delay until phase0 is right with 7M
    reg [1:0] state7M = 2'b00;
    reg bus_delay = 1'b0;
    
// bus delay state machine for device I/O   
// if you use 7M as input to the DCM then you must use: 
//  always @(posedge CLKIN_IBUFG_OUT) 

    always @(posedge BUS_7M)
        begin
        case (state7M)
        0: begin                                // start here
            if(BUS_PHASE0)              // check phase 0 
                begin                           // phase 0 is high
                state7M <= 1;               // next pass start with 1:
                bus_delay <= 1'b1;      // OK for device I/O
                end
            else bus_delay <= 1'b0;     // phase 0 is low - no device I/O - start over
            end
        1: begin
            if(BUS_PHASE0)              // check phase 0
                begin                           // phase 0 is high
                state7M <= 2;               // next pass start with 2:
                bus_delay <= 1'b1;      // OK for device I/O
                end
            else 
                begin                           // phase 0 is low
                state7M <= 0;               // start over
                bus_delay <= 1'b0;      // no device I/O
                end
            end
        2: begin
            if(BUS_PHASE0)              // check phase 0
                begin                           // phase 0 is high
                state7M <= 3;               // next pass start with 3:
                bus_delay <= 1'b1;      // OK for device I/O
                end
            else 
                begin                           // phase 0 is low
                state7M <= 0;               // start over
                bus_delay <= 1'b0;      // no device I/O
                end
            end
        3: begin
            state7M <= 0;                   // start over               
            bus_delay <= 1'b0;          // no device I/O
            end
        default: 
            begin
            state7M <= 0;                   // start over
            bus_delay <= 1'b0;          // no device I/O
            end
        endcase
        end
////////////////////////////////////////////////////////////////////////////////
// device_io 
// high when Apple /DEVSEL is low and phase 0 and BUS_7M are high per the 
// bus_delay state machine      
    wire device_io = (devsel_s == 1'b0) & bus_delay;
////////////////////////////////////////////////////////////////////////////////
// dev_read
// high when it's OK to read $C0nx area device I/O  
    wire dev_read = device_io & rw_s;   // device_io is high and rw_s (read/write) is high
////////////////////////////////////////////////////////////////////////////////
// dev_write
// high when it's OK to write $C0nx area device I/O 
    wire dev_write = device_io & ~rw_s; // device_io is high and rw_s (read/write) is low
`ifdef m25
////////////////////////////////////////////////////////////////////////////////
// flash chip interface
    reg write_spi_reg = 0;      // write to spi on = 1 for M25 read, program (slow) & sector erase
    reg write_spi_reg4 = 0;     // fastflash
    reg write_spi_reg6 = 0;     // read back error check
    reg [7:0] spi_in_reg = 0;   // registered spi input for 'read, program & sector erase' code
    reg [7:0] spi_in_reg4 = 0;  // registered spi input for fastflash
    reg [7:0] spi_in_reg6 = 0;  // registered spi input for readback - errors 
    wire one_byte_cmd;          // wire that is on for any one byte M25 command
    wire two_byte_cmd;          // wire that is on for any two byte M25 command         
    reg write_cmd_reg = 0;      // write to spi on = 1 for M25 1 & 2 byte commands
    reg [7:0] cmd_reg = 0;      // registered spi input for M25 1 & 2 byte commands 
`ifdef fastflash
    reg ff_on = 0;                  // fastflash on/off
    reg fr_on = 0;                  // fast read on/off
// write_spi
// determines when to write to spi from the various code that accesses it
    assign write_spi = (one_byte_cmd || two_byte_cmd) ? write_cmd_reg : (ff_on) ? write_spi_reg4 : fr_on ? write_spi_reg6 : write_spi_reg;
// spi_in
// get the byte to send to spi from the various code segments
    assign spi_in = (one_byte_cmd || two_byte_cmd) ? cmd_reg : ff_on ? spi_in_reg4 : fr_on ? spi_in_reg6 : spi_in_reg;
`else
// write_spi
// determines when to write to spi from the various code that accesses it
    assign write_spi = (one_byte_cmd || two_byte_cmd) ? write_cmd_reg : write_spi_reg;
// spi_in
// get the byte to send to spi from the various code segments
    assign spi_in = (one_byte_cmd || two_byte_cmd) ? cmd_reg : spi_in_reg;
`endif
    reg do_nvram = 0;                   // execution strobe
    reg cmd_cs = 1;                 // M25 on/off from 1 or 2 byte command process - on = 0
    reg spi_cs = 1;                 // M25 read, program (slow) & sector erase - M25 on = 0 off = 1
    reg spi_cs4 = 1;                    // fastflash page program
    reg spi_cs6 = 1;                    // read back error check
    reg [7:0] page_errs = 0;        // page error count
    reg [4:0] nvram_state6 = 0;     // read back error check 
////////////////////////////////////////////////////////////////////////////////
// create spi clock (clk25_b) and turn on/off M25 from any M25 routine
    always @(posedge clk50)
        begin
        clk25_b <= ~clk25_b;            // create spi clock
`ifdef fastflash
        if (clk25_b) spi_rom_cs <= (one_byte_cmd || two_byte_cmd) ? cmd_cs : ff_on ? spi_cs4 : fr_on ? spi_cs6 : spi_cs;
`else
        if (clk25_b) spi_rom_cs <= (one_byte_cmd || two_byte_cmd) ? cmd_cs : spi_cs;
`endif
        end
`endif //m25
////////////////////////////////////////////////////////////////////////////////
// Starting with bitstream version 9.20 for the Apple IIgs & version 3.00 for the Apple //e there
// is an I/O interface for sending or receiving data, cmds, etc.
// In version 9.22 there were new commands added for faster flashing.
////////////////////////////////////////////////////////////////////////////////
// The I/O locations:
// $C0n0 to C0nF where n = the slot# of the Carte Blanche + 8.
//
// $C0n0 is the command register. Writing a command number tells the CB how to use the rest of the I/O regs.
// Commands:
// 0 = does nothing but reset some of the other registers.
// 1 = puts CB code version in $C0n1 & $C0n2. 
// 2 = M25P80 NVRAM utilities (see below)
// 3 = does nothing
// 4 = FastFlash (see below)
// 5 = debug. The action of this command may change from version to version.
// 6 = Error check (see below)
// 7 = reserved
//
// $C0n1 is a data register.  It's use depends on the command reg.
// $C0n2 is a data register.  It's use depends on the command reg. In this version command 0 does not reset this register.
// $C0n3 use depends on command reg.
// $C0n4 address low
// $C0n5 address mid
// $C0n6 address high
// $C0n7 use depends on command reg.
// $C0n8 reserved -
// $C0n9 reserved    | - were used for freq. counter (may change)
// $C0nA reserved -
// $C0nC reserved
// $C0nF always $CB for ID.  Command 0 does not reset this register.  This can be used to find which slot the Carte Blanche is in.
`ifdef m25
////////////////////////////////////////////////////////////////////////////////
// M25P80 NVRAM I/O
// If $C0n0 = $02 then the M25P80 NVRAM I/O is active.
// $C0n1 = instruction
//  Write Enable = $06
//  Write Disable = $04 untested
//  Read Status  = $05
//  Read Data Bytes = $03
//  Page Program = $02
//  Sector Erase = $D8 (216)
//  Release from Deep Power Down = $AB (171) untested
//  Bulk Erase = $C7 (199) untested
//  Deep Power Down = $B9 (185) untested
//  Write Status = $01 (untested)
// $C0n2 = data byte to read or write
// $C0n3 = used as strobe to start execution of an instruction. All other registers should be set before strobing.
// $C0n4 = low byte of address in M25
// $C0n5 = mid byte of address in M25
// $C0n6 = high byte of address in M25.  The high 4 bits are ignored.
////////////////////////////////////////////////////////////////////////////////
// FastFlash
// Flashes one page (256 bytes) that have been loaded into $4000 - $40FF in Apple II memory.
// Use M25P80 NVRAM I/O instructions above to set Write Enable & address to write.
// The page must have been erased before it can be flashed.
//
// If $C0n0 = $04 then the FASTFLASH function is active.
// $C0n1 = instruction
//      Page Program = $02 (must be set for FASTFLASH)
// $C0n3 = used as strobe to start execution. (set $C0n0 & $C0n1 before strobing)
//////////////////////////////////////////////////////////////////////////////// 
// Error check
// Reads a page (256 bytes) found at $4000 - $40FF in Apple II memory and compares the bytes to those found at a specified address.
// Error check tests to make sure erasing or page programming is done before beginning.
// Use M25P80 NVRAM I/O instructions above to set address to check.
//
// If $C0n0 = $06 then the Error check function is active.
// $C0n1 = instruction 
//      Read Data Bytes = $03 (must be set for Error check)
// $C0n3 = used as strobe to start executionn. (set $C0n0 & $C0n1 before strobing)
// $C0n7 = returns the number of errors in the page.
`endif //m25
////////////////////////////////////////////////////////////////////////////////
// peripheral device I/O 
// used to communicate between Apple & CB
    reg [7:0] C0n0 = 0;         // command reg
    reg [7:0] C0n1 = 0;         // data reg (usually low byte)  
    reg [7:0] C0n2 = 0;         // data reg (usually high byte)
    reg [7:0] C0n3 = 0;         // start execution of M25 routine
    reg [7:0] C0n4 = 0;         // address low (M25)
    reg [7:0] C0n5 = 0;         // address mid  (M25)
    reg [7:0] C0n6 = 0;         // address high (M25)
    reg [7:0] C0n7 = 0;         // errors (M25)
    reg [7:0] C0nC = 0;         // not used - reserved
    reg [7:0] C0nF = 8'hCB; // ID byte read only
    reg [7:0] bus_reg = 0;
`ifdef m25
////////////////////////////////////////////////////////////////////////////////
// read_cmds - M25P80 
// read data bytes or read status
    wire read_cmds = (C0n1 == 8'h03) || (C0n1 == 8'h05);
`endif //m25
////////////////////////////////////////////////////////////////////////////////
// read and write device I/O
    always @(negedge Q3_s)  
        begin
`ifdef m25
        if (do_nvram == 1'b1) do_nvram <= 1'b0; // stay on for only one Q3 cycle
        if((spi_rom_cs == 1'b1))                        // M25P80 is on
            begin
            do_nvram <= 1'b0;                               // just in case we turn this off
            if ((C0n0 == 8'h02) && read_cmds) C0n2 <= spi_out;  // read back from M25
            if ((C0n0 == 8'h06))                            // error routine?
                begin                                           // yes
                    C0n7 <= page_errs;                  // return errors
                    C0n2 <= nvram_state6;               // where error occured in error test routine
                end                 
            end
`endif //m25                
// device I/O write
        if(dev_write)                           // test for device write
            begin                                   // find out what was written
            case(bus_a_s[3:0])
            4'h0:                                   // commands 0 = reset regs, 1 = version, 2 = M25P80, 
                                                    // 3 = do nothing, 4 = fastflash, 5 = test/debug, 6 = error test
                begin                               // only commands 0 & 1 are handled here
                C0n0 <= bus_d_s;            // make it stick
                if(bus_d_s == 8'h00)            // reset io regs command?
                    begin                           // yes, these are the only regs reset to zero
                    C0n1 <= 8'h00;              
                    C0n3 <= 8'h00;              // C0n2 is not reset
                    C0n4 <= 8'h00;
                    C0n5 <= 8'h00;
                    C0n6 <= 8'h00;
                    C0n7 <= 8'h00;
                    C0nC <= 8'h00;
                    end
                if(bus_d_s == 8'h01)            // version?
                    begin                           // yes
                    C0n1 <= min_version;        // 0 - 255
                    C0n2 <= maj_version;        // 0 - 254 (255 is used for test versions)
                    end                 
                end
            4'h1: C0n1 <= bus_d_s;          // data - depends on command (usually low byte) 
            4'h2: C0n2 <= bus_d_s;          // data - depends on command (usually high byte)
            4'h3:   begin
                    C0n3 <= bus_d_s;            // depends on command (starts execution for flashing & error check)
`ifdef m25
`ifdef fastflash                                // command reg must be set for M25 util., fastflash or error check
                    if((C0n0 == 8'h02) || (C0n0 == 8'h04) || (C0n0 == 8'h06))   
`else
                    if(C0n0 == 8'h02)           // command reg set M25 utilities 
`endif                  

                        begin
                        do_nvram <= 1'b1;       // start execution of M25 read/write
                        end
`endif //m25
                    end         
            4'h4: C0n4 <= bus_d_s;      // address low      
            4'h5: C0n5 <= bus_d_s;      // address mid
            4'h6: C0n6 <= bus_d_s;      // address high
            4'h7: C0n7 <= bus_d_s;      // page errors (fastflash)
            4'hC: C0nC <= bus_d_s;      // not used 
            endcase
            end
// device I/O read
        if(dev_read)
            begin                               // register data for reading
            case(bus_a_s[3:0])
            4'h0: bus_reg <= C0n0;      // read the command
            4'h1: bus_reg <= C0n1;      // read the data
            4'h2: bus_reg <= C0n2;      // read the data
            4'h3: begin
                    bus_reg <= C0n3;        // a read start execution same as a write
`ifdef m25
`ifdef fastflash                            // command reg must be set for M25 util., fastflash or error check
                    if((C0n0 == 8'h02) || (C0n0 == 8'h04) || (C0n0 == 8'h06)) do_nvram <= 1'b1;
`else
                    if(C0n0 == 8'h02) do_nvram <= 1'b1; // start execution of M25 read/write
`endif
`endif //m25
                    end
            4'h4: bus_reg <= C0n4;      // read address low
            4'h5:   bus_reg <= C0n5;        // read address mid
            4'h6: bus_reg <= C0n6;      // read address high
            4'h7: bus_reg <= C0n7;      // read page errors (fastflash)

            4'hC: bus_reg <= C0nC;      // not used
            4'hF:   bus_reg <= C0nF;        // always reads 'CB'
            endcase
            end     
        end
////////////////////////////////////////////////////////////////////////////////
// put register value on bus if device IO read else high-z      
    assign BUS_D = dev_read ? bus_reg : 8'bz;   
`ifdef m25
////////////////////////////////////////////////////////////////////////////////    
// Device IO command $02 (M25P80 active)
// read, program or sector erase M25 
// slow - one byte at a time program (flash) 
// Read Data Bytes (READ)   $03
//  Page Program     (PP)       $02
//  Sector Erase     (SE)       $D8
// clk25_b is half the main clock

    reg [4:0] nvram_state = 0;

    always @(posedge clk25_b or negedge reset_s)
        begin
// reset
        if(reset_s == 1'b0)                         // test for reset
            begin
            spi_cs <= 1'b1;                         // turn off M25
            write_spi_reg <= 1'b0;                  // writing is off
            nvram_state <= 0;                           // reset state
            end
// read - sector erase - slow flash (one byte at a time) if M25P80 active   
        else if(((C0n1 == 8'h03) || (C0n1 == 8'h02) || (C0n1 == 8'hD8)) && (C0n0 == 8'h02))    
            begin
            case (nvram_state)
            0: begin                                        
                if(do_nvram)                            // execution strobe?
                    begin                                   // yes
                    nvram_state <= 1;                   // next state
                    spi_cs <= 1'b0;                 // turn on M25
                    spi_in_reg <= C0n1;             // send read, program or sector erase command to spi
                    write_spi_reg <= 1'b1;          // write command
                    end
                end
            1: nvram_state <= 2;                        // next state - delay one cycle
            2: begin                                        
                write_spi_reg <= 1'b0;              // stop writing     
                if (spi_done)                           // spi routine done?  if not keep testing until it is
                    begin                                   // yes it's done
                    nvram_state <= 3;                   // next state
                    end
                end
            3: begin                                        
                spi_in_reg <= C0n6;                 // send high addr byte to spi
                nvram_state <= 4;                       // next state
                write_spi_reg <= 1'b1;              // write byte
                end
            4: nvram_state <= 5;                        // next state - delay
            5: begin                                        
                write_spi_reg <= 1'b0;              // stop writing
                if (spi_done)                           // spi routine done?  if not keep testing until it is
                    begin                                   // yes it's done
                    nvram_state <= 6;                   // next state
                    end
                end
            6: begin                                    
                spi_in_reg <= C0n5;                 // send mid addr byte to spi
                nvram_state <= 7;                       // next state
                write_spi_reg <= 1'b1;              // write byte
                end
            7: nvram_state <= 8;                        // next state - delay
            8: begin                                        
                write_spi_reg <= 1'b0;              // stop writing
                if (spi_done)                           // spi routine done?  if not keep testing until it is 
                    begin                                   // yes it's done
                    nvram_state <= 9;                   // next state
                    end
                end
            9: begin                            
                spi_in_reg <= C0n4;                 // send low addr byte to spi
                nvram_state <= 10;                  // next state       
                write_spi_reg <= 1'b1;              // write byte
                end
            10: nvram_state <= 11;                  // next state - delay
            11:begin                                    
                write_spi_reg <= 1'b0;              // stop writing
                if (spi_done)                           // spi routine done? if not keep testing until it is
                    begin                                   // yes it's done
                    if (C0n1 == 8'hD8)              // are we doing a sector erase?
                        begin                               // yes
                        spi_cs <= 1'b1;             // turn off M25
                        nvram_state <= 0;               // reset state  - we're done with sector erase
                        end
                    else                                    // no it's not a sector erase
                        begin                               
                        spi_in_reg <= (C0n1 == 8'h02) ? C0n2 : 8'h00;   // if page program send a byte to spi
                                                                                        // if not send a zero 
                        nvram_state <= 12;          // next state
                        write_spi_reg <= 1'b1;      // write byte
                        end
                    end
                end
            12: nvram_state <= 13;                  // next state - delay
            13:begin
                write_spi_reg <= 1'b0;              // stop writing     
                if (spi_done)                           // spi routine done? if not keep testing until it is
                    begin                                   // yes, we're done
                    spi_cs <= 1'b1;                 // turn off M25
                    nvram_state <= 0;                   // reset state 
                    end
                end
            default: ;                                  // do nothing
            endcase
            end
        end
`endif // m25       
////////////////////////////////////////////////////////////////////////////////        
// Fast Flash   
// uses page program to flash 256 bytes from SRAM 
`ifdef fastflash
    reg [4:0] nvram_state4 = 0; 

always @(posedge clk25_b or negedge reset_s)
    begin
// reset
    if(reset_s == 1'b0)                         // reset?
        begin                                           // yes
        spi_cs4 <= 1'b1;                            // turn off M25
        write_spi_reg4 <= 1'b0;                 // turn off writing
        nvram_state4 <= 0;                      // reset state to zero
        ff_addr <= 9'h000;                      // reset fastflash data address to $000
        ff_on  <= 1'b0;                         // fastflash is off
        end
// not reset - flash
    else if  ((C0n1 == 8'h02) && (C0n0 == 8'h04))   // fast flash using page program
        begin
        case (nvram_state4)     
        0: begin                                            
            if(do_nvram)                            // execution strobe?                            
                begin                                   // yes
                ff_on <= 1'b1;                      // turn on fastflash
                nvram_state4 <= 1;              // next state
                spi_cs4 <= 1'b0;                    // turn on M25
                spi_in_reg4 <= C0n1;                // send command to spi (program only) 
                write_spi_reg4 <= 1'b1;         // write command 
                end
            end
        1: nvram_state4 <= 2;                   // next state - delay   
        2: begin                                        
            write_spi_reg4 <= 1'b0;             // stop writing 
            if (spi_done)                           // spi routine done? if not keep testing until it is                            
                begin                                   // yes
                nvram_state4 <= 3;              // next state
                end
            end
        3: begin                                        
            spi_in_reg4 <= C0n6;                    // send high addr byte to spi
            nvram_state4 <= 4;                  // next state
            write_spi_reg4 <= 1'b1;             // write byte
            end
        4: nvram_state4 <= 5;                   // next state - delay           
        5: begin                                        
            write_spi_reg4 <= 1'b0;             // stop writing 
            if (spi_done)                           // spi routine done? if not keep testing until it is     
                begin                                   // yes
                nvram_state4 <= 6;              // next state
                end
            end
        6: begin                                    
            spi_in_reg4 <= C0n5;                    // send mid addr byte to spi
            nvram_state4 <= 7;                  // next state
            write_spi_reg4 <= 1'b1;             // write byte
            end
        7: nvram_state4 <= 8;                   // next state - delay
        8: begin                                        
            write_spi_reg4 <= 1'b0;             // stop writing
            if (spi_done)                           // spi routine done? if not keep testing until it is 
                begin                                   // yes
                nvram_state4 <= 9;              // next state
                end
            end
        9: begin                            
            spi_in_reg4 <= 8'h00;               // send low addr byte (always zero) to spi
            nvram_state4 <= 10;                 // next state       
            write_spi_reg4 <= 1'b1;             // write byte
            ff_addr <= 9'h000;                  // set data address to $000
            end
        10: begin
             nvram_state4 <= 11;                    // next state - delay
             end
        11:begin    
            write_spi_reg4 <= 1'b0;             // stop writing
            if (spi_done)                           // spi routine done? if not keep testing until it is 
                begin                                   // yes - this is where we flash 256 bytes 
                spi_in_reg4 <= ff_array[ff_addr[7:0]];  // send byte to spi 
                ff_addr <= ff_addr + 1;         // increment address in array
                if (ff_addr == 9'h100)          // done incrementing? if not write another byte
                    begin                               // yes
                    nvram_state4 <= 12;         // next state
                    ff_addr <= 9'h000;              // reset array address      
                    end
                else nvram_state4 <= 10;        // no, not done incrementing 
                write_spi_reg4 <= 1'b1;         // write byte           
                end
            end
        12:begin
            nvram_state4 <= 13;                 // next state - delay
            end
        13:begin
            write_spi_reg4 <= 1'b0;             // stop writing
            if (spi_done)                           // spi routine done? if not keep testing until it is                        
                begin                                   // yes      
                spi_cs4 <= 1'b1;                    // turn off M25
                nvram_state4 <= 0;              // reset state
                ff_on  <= 1'b0;                 // fastflash is off
                end
            end
        default: ;                                  // do nothing
        endcase
        end
    end 
////////////////////////////////////////////////////////////////////////////////        
// read back & test for errors 
always @(posedge clk25_b or negedge reset_s)
    begin
// reset
    if(reset_s == 1'b0)                         // reset?
        begin                                           // yes
        spi_cs6 <= 1'b1;                            // turn off M25                         
        write_spi_reg6 <= 1'b0;                 // turn off writing
        nvram_state6 <= 0;                      // reset state to zero
        fr_addr <= 9'h000;                      // reset read addres
        fr_on  <= 1'b0;                         // turn off readback
        page_errs <= 8'h00;                     // reset errors
        end
// not reset - read and error check
    else if  ((C0n1 == 8'h03) && (C0n0 == 8'h06))
        begin
        case (nvram_state6) 
        0:  begin
            if(do_nvram)                            // execution strobe?    
                begin                                   // yes
                fr_on <= 1'b1;                      // turn on readback
                fr_addr <= 9'h000;              // start with address $000
                nvram_state6 <= 1;              // next state
                spi_in_reg6 <= 8'h05;           // send read status command to spi 
                write_spi_reg6 <= 1'b1;         // write it
                spi_cs6 <= 1'b0;                    // turn on M25
                end
            end
        1: nvram_state6 <= 2;                   // next state - delay
        2: begin                                        
            write_spi_reg6 <= 1'b0;             // stop writing             
            if (spi_done) nvram_state6 <= 3;    // spi routine done? if not keep testing until it is 
            end                                     // yes - next state
        3: begin
            spi_in_reg6 <= 8'h00;               // send dummy byte to spi
            nvram_state6 <= 4;                  // next state
            write_spi_reg6 <= 1'b1;             // write byte
            end
        4: nvram_state6 <= 5;                   // next state - delay
        5: begin                                        
            write_spi_reg6 <= 1'b0;             // stop writing             
            if (spi_done)                           // spi routine done? if not keep testing until it is  
                begin                                   // yes
                if(spi_out[0] == 1'b0)          // low bit of spi output = 0?
                    begin                               // yes
                    spi_cs6 <= 1'b1;                // turn off M25 
                    nvram_state6 <= 6;          // next state
                    page_errs <= 8'h00;         // no error
                    end
                else                                    // low bit of spi output = 1        
                    begin
                    nvram_state <= 1;               // go back to state 1 
                    spi_in_reg6 <= 8'h05;       // send read status to spi 
                    write_spi_reg6 <= 1'b1;     // write it         
                    end
                end
            end
        6: begin
            spi_in_reg6 <= C0n1;                    // send read command to spi
            nvram_state6 <= 7;                  // next state
            spi_cs6 <= 1'b0;                        // turn on M25
            write_spi_reg6 <= 1'b1;             // write command
            end
        7: nvram_state6 <= 8;                   // next state - delay
        8: begin                                        
            write_spi_reg6 <= 1'b0;             // stop writing             
            if (spi_done) nvram_state6 <= 9;    // spi routine done? if not keep testing until it is            
            end                                     // yes - next state
         9:begin                                        
            spi_in_reg6 <= C0n6;                    // send high addr byte to spi
            nvram_state6 <= 10;                 // next state
            write_spi_reg6 <= 1'b1;             // write byte
            end
        10: nvram_state6 <= 11;                 // next state - delay
        11:begin                                        
            write_spi_reg6 <= 1'b0;             // stop writing
            if (spi_done) nvram_state6 <= 12; // spi routine done? if not keep testing until it is 
            end                                     // if so next state
        12:begin                                    
            spi_in_reg6 <= C0n5;                    // send mid addr byte to spi
            nvram_state6 <= 13;                 // next state                   
            write_spi_reg6 <= 1'b1;             // write byte
            end
        13:nvram_state6 <= 14;                  // next state - delay
        14:begin                                        
            write_spi_reg6 <= 1'b0;             // stop writing
            if (spi_done) nvram_state6 <= 15;   // spi routine done? if not keep testing until it is 
            end                         
        15:begin                            
            spi_in_reg6 <= fr_addr[7:0];        // send low addr byte to spi
            nvram_state6 <= 16;                 // next state           
            write_spi_reg6 <= 1'b1;             // write byte
            end
        16:nvram_state6 <= 17;                  // next state - delay
        17:begin
            write_spi_reg6 <= 1'b0;             // stop writing     
            if (spi_done)                           // spi routine done? if not keep testing until it is 
                begin
                nvram_state6 <= 18;             // next state           
                write_spi_reg6 <= 1'b1;         // write ????  why? 
                end
             end
        18:nvram_state6 <= 19;                  // next state - delay
        19:begin    
            write_spi_reg6 <= 1'b0;             // stop writing     
            if (spi_done)                           // spi routine done? if not keep testing until it is 
                begin                                   // check for error
                if(spi_out != ff_array[fr_addr[7:0]]) page_errs <= page_errs + 1; // if error then increment error count
                fr_addr <= fr_addr + 1;         // increment address
                if(fr_addr == 9'h100)           // checked 256 bytes?
                    begin                               // yes
                    fr_addr <= 9'h000;          // reset address
                    nvram_state6 <= 20;         // next state (last)
                    end
                else                                    // no not all 256 checked
                    begin
                    nvram_state6 <= 6;          // go back to state 6 to get another byte
                    spi_cs6 <= 1'b1;                // turn off M25
                    end
                end
            end
        20:begin
                begin
                spi_cs6 <= 1'b1;                    // turn off M25
                nvram_state6 <= 0;              // reset state to zero
                fr_on <= 1'b0;                      // turn off readback
                end
            end
        default: ;                                  // do nothing
        endcase
        end
    end
`endif // fastflash
`ifdef m25
////////////////////////////////////////////////////////////////////////////////
// Process one or two byte commands (M25)
// Some M25P80 commands require just one or two bytes 
//
// -one byte NVRAM commands-
// Write enable (WREN)  $06
//  Write disable (WRDI)    $04
//  Bulk Erase (BE)     $C7
//  Deep Power Down (DP)    $B9
//  Release Deep Power Down (RES)   $AB
//
// -two byte NVRAM commands-
// Read Status Register (RDSR)  $05
//  Write Status Register (WRSR)    $01

reg [2:0] cmd_state = 0;

assign one_byte_cmd = (C0n1 == 8'h06) || (C0n1 == 8'h04) || (C0n1 == 8'hC7) || (C0n1 == 8'hB9) || (C0n1 == 8'hAB) && (C0n0 == 8'h02);
assign two_byte_cmd = (C0n1 == 8'h05) || (C0n1 == 8'h01) && (C0n0 == 8'h02);
// Process commands
    always @(posedge clk25_b or negedge reset_s)
        begin
// reset
        if(reset_s == 1'b0)                     // reset?
            begin                                       // yes
            cmd_cs <= 1'b1;                     // M25 off
            cmd_state <= 0;                     // reset state
            write_cmd_reg <= 1'b0;              // writing off
            end
// not reset
        else if((C0n0 == 8'h02) && (one_byte_cmd || two_byte_cmd))
            begin
            case (cmd_state)
            0: begin                                    // use 1 cycle
                if (do_nvram)                       // execution strobe?
                    begin                               // yes
                    cmd_state <= 1;             // next state
                    cmd_cs <= 1'b0;             // turn on M25
                    cmd_reg <= C0n1;                // send a one or two byte command to spi
                    write_cmd_reg <= 1'b1;      // write command
                    end
                end
            1: cmd_state <= 2;                  // next state - delay
            2: begin
                write_cmd_reg <= 1'b0;          // stop writing
                if (spi_done)                       // spi routine done? if not keep testing until it is                        
                    begin                               // yes 
                    if (one_byte_cmd)               // one byte command?
                        begin                           // yes
                        cmd_cs <= 1'b1;         // turn off M25 we're done
                        cmd_state <= 0;         // reset state
                        end
                    else                                // it's a 2 byte command
                        begin                           // continue if reading status reg
                        cmd_reg <= 8'h02;           // dummy needed for 2 byte commands
                        cmd_state <= 3;         // next state
                        write_cmd_reg <= 1'b1;  // write the dummy byte
                        end
                    end
                end
            3: cmd_state <= 4;                  // next state - delay
            4: begin
                write_cmd_reg <= 1'b0;          // stop writing
                if (spi_done)                       // spi routine done? if not keep testing until it is                        
                    begin                               // yes
                    cmd_cs <= 1'b1;             // turn off M25
                    cmd_state <= 0;             // reset state
                    end
                end
            default: ;                              // do nothing
            endcase
            end
        end
`endif // m25
////////////////////////////////////////////////////////////////////////////////                
// Synchronize Apple slot clocks to CB clock
// run each signal or clock through two flip-flops
    always @(posedge clk50)     
        begin

        if(~LOCKED_OUT)     // if DCM is not up and running
            begin                   // initial settings for the registers
    //      reset_s <= 1;       // this happens only at startup
            phase0_s <= 0;  
            Q3_s <= 1;      
            devsel_s <= 1;  
            m2b0_s <= 0;        
            bus_a_1 <= 0;   
            bus_a_s <= 0;   
            bus_d_1 <= 0;       
            bus_d_s <= 0;
            user1_s <= 0;
            ph1_s <= 1; 
            rw_s <= 1'bz;
            bus_7M_s <= 1;
            iosel_s <= 1;
            rw_1 <= 1'bz;
    //      reset_1 <= 0;
            phase0_1 <= 0;
            Q3_1 <= 1;
            devsel_1 <= 1;
            m2b0_1 <= 0;
            ph1_1 <= 1;
            user1_1 <= 0;
            bus_7M_1 <= 1;
            iosel_1 <= 1;           
            end
        else                        // DCM is up and running
            begin                   // sync with 50MHz clock
            rw_1 <= BUS_R_W;
            rw_s <= rw_1;
            Q3_1 <= BUS_Q3;
            Q3_s <= Q3_1;
            phase0_1 <= BUS_PHASE0;
            phase0_s <= phase0_1;
            ph1_1 <= BUS_PHASE1;
            ph1_s <= ph1_1;
            reset_1 <= BUS_RESET;
            reset_s <= reset_1;
            devsel_1 <= N_BUS_DEV_SEL;
            devsel_s <= devsel_1;
            m2b0_1 <= BUS_3M58;
            m2b0_s <= m2b0_1;
            bus_a_1 <= BUS_A;
            bus_a_s <= bus_a_1;
            bus_d_1 <= BUS_D;
            bus_d_s <= bus_d_1;
            user1_1 <= BUS_USER1;
            user1_s <= user1_1;
            bus_7M_1 <= BUS_7M;
            bus_7M_s <= bus_7M_1;
            iosel_1 <= N_BUS_IO_SEL;
            iosel_s <= iosel_1;
            end
        end
////////////////////////////////////////////////////////////////////////////////    
// bugin
// allows you to send a one bit state of any wire or flip-flop to vga_video.v
// where it will increment the border color # by it's value 
// 0 - normal border   1 - increment the color
`ifdef bbug
    assign bugin = reset;   // replace 1'b0 with a wire/reg to send state to bugborder
`endif 
endmodule
////////////////////////////////////////////////////////////////////////////////    

