//////////////////////////////////////////////////////////////////////////////////
// vga_video.v  -  c933hdmi.bit
// This is a module for the Carte Blanche (CB) card under the top module cb1.v.
// Its main functions are:
//
// Generate video output as VGA and/or HDMI.
//  It does this by maintaining counters for horizontal and vertical positions 
//  of the visible pixels and the blanking.  The horizontal and vertical sync
//  pulses are also determined by those counters.
//
//  It creates the pixels for display by processing the data from buffers in 
//  the CB SRAM and/or FPGA BRAM. The video differs depending on the video 
//    mode. The resulting pixel data is sent to cb1.v for output.
//
//  It determines the correct video mode (text or various graphics modes)
//      by watching the Apple slot bus for reads/writes to the Apple II softswitch
//      area at $C0xx.
// 
// Determine the correct timing for reading from or writing to the video buffers.
// 
// SHR 640 "dithered" mode. 
// Keypress input routine for testing and toggling "dithered" SHR 640 mode
// by pressing control-] 
//////////////////////////////////////////////////////////////////////////////////
// This version is for hdmi using the hdmi/dvi peripheral board for the CB.
// It is for an Apple IIgs only.  
// The 12-bit mod is not supported.  6-bit will still be available 
// The resolution is 720x480.  The DCM (main clock) output must be ~56.644 MHz 
// Signals or clocks that are synchronized with the main clock appended with _s
//////////////////////////////////////////////////////////////////////////////////
`include "c933hdmi_defs.vh" // contains global definitions

module vga_video (                  
          input clk50,              // ~57.2727MHz from DCM (higher than 56.644 MHz it should be)
          input [7:0] char_data,    // from font (distributed RAM is used instead of block RAM)
          input [7:0] ram_data,     // from video data SRAM thru cb1.v
          input [15:0] bus_addr,    // Apple slot bus address - sync'd in cb1.v
          input [7:0] bus_data,     // Apple slot bus data - sync'd in cb1.v
          input ph1_s,                  // Apple phase1 clock - sync'd in cb1.v
          input phase0_s,               // Apple phase0 clock 
          input Q3,                     // Apple Q3 clock - sync'd in cb1.v
//        input rw,                     // read/write from Apple
          input rw_s,                   // read/write from Apple - sync'd in cb1.v
          input reset_n,                // Apple reset inverted (high when reset)
          input user1_s,                // M2SEL (for GS) on //e pin 39 - sync'd in cb1.v
          input devsel_s,               // Apple slot /DEVSEL               
          input m2b0_s,             // aka 3M58 is only available on slot 3 on a ROM 01 gs.  1-6 on a ROM 03
`ifdef bbug
            input bugin,                // use this to pass a signal for bugborder
`endif  
            input [7:0] scb_data,       // input scanline from BRAM
            output [7:0] scb_addr,      // output scanline address in BRAM
            input [11:0] pal_data,      // input palette color from BRAM
            output [7:0] pal_addr,      // output palette color address in BRAM
            output update_time,         // when it's OK to write Apple data to SRAM
            output reg auxmem,          // high when mirroring auxram (bank1) to CB RAM 
            output video_read,          // high when it's OK to read the CB RAM 
            output v_sync,                  // 6 bit & 24 bit hdmi
            output h_sync,                  // 6 bit & 24 bit hdmi
// --- hdmi/dvi PB --- //
            output [7:0] RED,               // 8 bits red
            output [7:0] GREEN,         // 8 bits green
            output [7:0] BLUE,          // 8 bits blue
            output FRAME_ENABLE,            // high when pixels are drawn
            output PIX_CLK,         // hdmi pixel clock
            output COLOR_LEVEL,         // 1 = 24 bit, 0 = 12 bit hdmi
// -------------------- //
            output reg [10:0] rom_addr,     // address used to get character data
            output [16:0] ram_address,  // address to get ram_data from CB RAM buffer
`ifdef tester
          output [9:0] wtest00,         // 0 - 1023 register used for debug testing
`endif
          output clk25,                 // half of clk50 - for 720x480 should be ~28.322 MHz
          output [5:0]  RGB             // 6 bit RGB used for the original CB 6-bit VGA connector
          );
///////////////////////////////////////////////////////////////////////////
// registers used for synchronizing Apple signals with the main CB clock //
    reg reset_s = 0;
    reg reset_1 = 0;
    reg m2b0_sl = 0;
///////////////////////////////////////////////////////////////////////////
// the two main counters for positions of pixels, blanking and sync signals
    reg [9:0] h_count = 0;  // horz dot counter 
   reg [9:0] v_count = 47;  // vert line counter
// pixel clock  
   reg clk1 = 0;                    // half frequency of main clock
///////////////////////////////////////////////////////////////////////////
// settings from softswitch area $C0xx
   reg TEXT = 1;                // TEXT mode initially on
   reg MIXED = 1;               // MIXED mode initially on (4 lines of text)
   reg PAGE2 = 0;               // PAGE 2 initially off
    reg HIRES = 0;              // High Res graphics initially off
   reg mode_80COL = 0;      // 80 column mode initially off
    reg AltCharSet = 0;     // alternate character set initially off
    reg mode_80STORE = 0;   // 80 column mode initially off
    reg WRITERAM_SEL = 0;   // 1 = aux-same as $C005 on, 0 = main - same as $C004 on
    reg LINEAR_SHR = 0;     // $C029 bit 6
    reg SHR = 0;                // Super Hires graphics initially off               
    reg DHR = 0;                // Double Hires graphics initially off
    reg monocolor = 0;      // monocolor Double Hires graphics initially off
    reg [3:0] BORDER_COLOR_NUM = 4'b0000;   // border init to black
    reg [3:0]   TEXT_COLOR_NUM = 4'b1111;       // text init to white   
    reg [3:0]   BACK_COLOR_NUM = 4'b0000;       // background init to black
    reg [7:0] CBVID = 8'h00;    // CB video switch $C06A
    reg option_key = 1'b0;  // 1 if key pressed
///////////////////////////////////////////////////////////////////////////
// testing routines used for debugging - ifdef'ed out when code released
// bugborder uses VGA border color as data tester   
`ifdef bbug
    reg bugborder = 0;      // added 9/30/10 for testing wires & single bit regs
`endif
///////////////////////////////////////////////////////////////////////////
// used for various code testing
`ifdef tester
    reg[7:0] KEY  = 0;      // keypress reg initialized to 0    
    reg[7:0] test42 = 0;        // input for various tests
    reg[9:0] test00 = 0;        // counter for tests using keypresses
    assign wtest00 = test00;    // for passing to cb1.v
`endif
///////////////////////////////////////////////////////////////////////////
    reg [7:0]   shifter7 = 8'h00;               // shifter used in non-super hires modes 
   wire [2:0]   scanline = v_count[3:1];    // used for determining video address
   reg [2:0]    count7 = 3'b000;                // 0-6 counter used with shifter7
   reg [6:0]    hc = 0;                         // horizontal character count
   reg [4:0]    vc = 0;                         // vertical character count
    reg [4:0]   count28 = 0;                    // 0-27 counter added for DHR
    reg [7:0]   shr_bytes = 0;              // super hires bytes in 1 scanline  
   reg [15:0] shr_addr = 16'h2000;          // address of bytes for super hires - 1
    wire disp40_active;                         // on when visible part of display is active
    wire border_active;                         // on when border is displayed
    wire shift_bitsum = disp40_active ? shifter7 > 0 : 1'b0;    // used with Hires 
   wire GR =  ~(TEXT | ( MIXED &  vc[4] & vc[2]));              // not text or mixed and >= line 20 
   reg GR_r = 0, GR_r2 = 0;                 // used to delay graphics
    wire hgr = (HIRES & GR_r2);             // HIRES delayed
   reg hor_active_reg = 0;                          // gets set to hor_active (wire) 
///////////////////////////////////////////////////////////////////////////
// colors_6 and colors_12 submodules (16 colors)
// registers are inputs - wires are outputs
// HIRES and Super HIRES do not use these submodules
// Low Resolution   
   reg [3:0]    low_color_nibble = 0;       // used to select lores color 
   wire [11:0] lowres_color_val12;  // actual 12 bit lores color
   wire [5:0]   lowres_color_val6;  // actual 6 bit lores color
// border, text & background    
    reg [3:0] border_number = 0;            // used to select border color 
    wire [5:0] border_color6;           // actual 6 bit color (border)
    wire [11:0] border_color12;     // actual 12 bit color (border)
    reg [3:0] text_number = 0;              // used to select text color 
    wire [5:0] text_color6;             // actual 6 bit color (text)
    wire [11:0] text_color12;           // actual 12 bit color (text)
    reg [3:0] back_number = 0;              // used to select background color
    wire [5:0] back_color6;             // actual 6 bit color (background)
    wire [11:0] back_color12;           // actual 12 bit color (background)
// Double High Resolution
    reg [3:0] dhrcolor_number = 0;          // used to select double hires color 
    wire [5:0] dhr_color6;              // double hires 6 bit color
    wire [11:0] dhr_color12;            // double hires 12 bit color
// Double Low Resolution
    wire[3:0] dlr_nibble = low_color_nibble;  // dlr uses dhr palette
    wire [11:0] dlr_color12;            // double lores 12 bit color
/////////////////////////////////////////////////////////////////////////// 
    wire dhires;                            // double hires, graphics and 80 columns on 
    wire shires;                            // linear super hires on 
    reg [5:0] color6 = 0;                       // hires 6 bit color
    reg [11:0] color12 = 0;                 // hires 12 bit color
    reg [11:0] grayscale = 0;               // 12 bit grayscale 
    wire [5:0] grayscale6 = {grayscale[11:10], grayscale[7:6], grayscale[3:2]}; // 6 bit grayscale
   reg hor_del2 = 0;                            // delays displaying horizontal pixel data
   reg [4:0] flash_count = 0;               // counter for FLASH text speed
   reg flash = 0;                               // FLASH text mode
    reg inverse = 0;                            // INVERSE text mode
   wire flashing = flash ? flash_count[4] : 1'b0; // FLASH on and off     
    
    wire Q3_s = Q3;                     // wire Q3 synchronized with main clock from cb1.v
   reg vid_del = 0;                         // video delay  
    assign clk25 = clk1;                    // send clk1 to cb1.v
    assign dhires = DHR & mode_80COL & GR;  // double hires, 80 column and graphics on
    assign shires = SHR & LINEAR_SHR;       // super hires  linear mode
//////////////////////////////////////////////////////////////////////////////////  
// blanking 
    wire h_blank = (h_count > 719);     // horz blanking
   wire v_blank = (v_count > 479);  // vertical blanking
//////////////////////////////////////////////////////////////////////////////////
// HDMI needs a frame enable signal that is high when pixels are to be drawn
    reg frame_enable_reg = 0;
    always @(*)
//  always @(posedge clk50)
        begin
        /*if (clk1)*/ frame_enable_reg = ~((h_count > 719) | (v_count > 479));
        end
        
// HDMI als needs color level signal        
    assign COLOR_LEVEL = 1'b1;   // 1= 24 bit, 0 = 12 bit       
//////////////////////////////////////////////////////////////////////////////////  
// create h_sync & v_sync
// h_sync values are for 720x480 resolution v_sync same as 640x480 resolution       
    assign      h_sync = (h_count < 741) || (h_count > 849); // 659 & 755 for 640 wide
    assign      v_sync = (v_count < 493) || (v_count > 494); // 493 & 494 is std.
//////////////////////////////////////////////////////////////////////////////////  
// determine when the visible display is active
    reg h_active = 0;   // horizontal pixels being displayed
    reg v_active = 0; // lines being displayed

    always @(posedge clk50)
        begin
        h_active <= (SHR == 1) ? ((h_count > 39) & (h_count < 680)) : ~mode_80COL ? (h_count > 54 ) & (h_count < 629) :  (h_count > 61) & (h_count < 629) ;     
        v_active <= (SHR == 1) ? (v_count > 47) & ( v_count < 448) : (v_count > 47) & ( v_count < 432); 
        end

    wire hor_active = h_active;
    wire vert_active = v_active;

    assign disp40_active =  SHR ? (h_active & v_active) : hor_active_reg & hor_del2 & vert_active;  // pixels, lines and delay
//////////////////////////////////////////////////////////////////////////////////
// border active
    assign border_active = (h_count < 720) & (v_count < 480) & ~disp40_active;
//////////////////////////////////////////////////////////////////////////////////  
// determine when it's OK to read and write RAM buffers (in cb1.v)
// update_time is for writing to the buffers    
    assign update_time = (hgr & ~DHR) ?  ((count7 == 0) & (h_count[0] == 1'b1)) /* : DHR ? (h_count[1] == 1'b1) */ : (h_count[1] == 1'b0); 

//  assign update_time = (SHR) ? (h_count[1:0] == 2'b00) : (HIRES /*hgr*/ & ~DHR) ?  ((count7 == 0) & (h_count[0] == 1'b1))  : (DHR) ? (h_count[1] == 1'b1)  : (h_count[1] == 1'b0); 
// video_read is for reading from the buffers
    assign video_read = 1'b1; // SRAM datasheet says /OE can be low for both read and write
//////////////////////////////////////////////////////////////////////////////////
// RGB and HDMI sent to cb1.v module
// vp used with both RGB & HDMI
    wire vp =  (FRAME_ENABLE) & (shifter7[0] ^ (inverse ^ flashing));   // TEXT bits
//  wire vp =  ~(h_blank | v_blank) && (shifter7[0] ^ (inverse ^ flashing));    // TEXT bits
//////////////////////////////////////////////////////////////////////////////////
// RGB 6 bit    
    wire [11:0] hdmi;   // 12 bit HDMI color used to create 6 bit color
    wire [5:0] GR_RGB = (dhires && monocolor && HIRES) ? grayscale6 : (dhires && HIRES) ? dhr_color6 : (HIRES == 0) ? lowres_color_val6 : color6; 

    assign RGB = (h_blank | v_blank) ? 6'b0 : (~disp40_active) ? border_color6 : (SHR) ? {(hdmi[11] & hdmi[10]), 
            (hdmi[9] & hdmi[8]), (hdmi[7] & hdmi[6]), (hdmi[5] & hdmi[4]), (hdmi[3] & hdmi[2]), (hdmi[1] & hdmi[0])} : 
            (GR_r2) ? GR_RGB : (({vp,vp,vp,vp,vp,vp} & text_color6[5:0]) ^ ({~vp,~vp,~vp,~vp,~vp,~vp} & back_color6[5:0]));
//////////////////////////////////////////////////////////////////////////////////
// colors_6.v submodule instanstiation (6 bit color tables)
    colors_6 colors_6 (
        .border_number(border_number),      // input
        .text_number(text_number),              // input
        .back_number(back_number),              // input
        .dhrcolor_number(dhrcolor_number),  // input    
        .nibble(low_color_nibble),              // input
        .border_color6(border_color6),      // output
        .text_color6(text_color6),              // output
        .back_color6(back_color6),              // output
        .dhr_color6(dhr_color6),                // output
        .RGB_val(lowres_color_val6)         // output
            );
//////////////////////////////////////////////////////////////////////////////////          
// HDMI / DVI Peripheral board video through cb1.v
    reg[11:0] hdmi2 = 0;                    // 12 bit color used for both 24 bit hdmi and 6 bit VGA
    reg[11:0] shr_color12 = 0;          // 12 bit Super Hires color
// TEXT color (12 bit) with color background
    wire [11:0] text12 = (({vp,vp,vp,vp,vp,vp,vp,vp,vp,vp,vp,vp} & text_color12[11:0]) ^ ({~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp} & back_color12[11:0]));
//  Graphics color (12 bit) not including SHR
    wire [11:0] GR_RGB12 = (dhires && monocolor && HIRES) ? grayscale : (dhires && HIRES) ? dhr_color12 : (HIRES == 0) ? lowres_color_val12 : color12; 
    assign FRAME_ENABLE = (h_count < 640) & (v_count < 480);
    assign PIX_CLK = clk1;                          // pixel clock for hdmi
    assign hdmi = hdmi2;                                // let 6 bit VGA use the same colors as hdmi

// when drawing to screen (frame enabled) get the colors depending on the video mode
// doing it when the pixel clock is low (~clk1) removes unwanted SHR screen right shift
    always @(posedge clk50)
        begin
        if (~clk1 & FRAME_ENABLE) hdmi2 <= (border_active) ? border_color12 : (SHR) ? shr_color12 : (GR_r2) ? GR_RGB12 : text12 ;
        end                                             //
// convert the 12 bit colors to 24 bits for hdmi output by repeating each color's 4 bit sequence
    assign RED = {hdmi2[11],hdmi2[10],hdmi2[9],hdmi2[8],hdmi2[11],hdmi2[10],hdmi2[9],hdmi2[8] };
    assign GREEN = {hdmi2[7],hdmi2[6],hdmi2[5],hdmi2[4],hdmi2[7],hdmi2[6],hdmi2[5],hdmi2[4]};
    assign BLUE = {hdmi2[3],hdmi2[2],hdmi2[1],hdmi2[0],hdmi2[3],hdmi2[2],hdmi2[1],hdmi2[0]};
//////////////////////////////////////////////////////////////////////////////////
// colors_12.v submodule instanstiation (12 bit color tables)
    colors_12 colors_12 (
`ifdef tester
        .test00(test00),
        .test42(test42),
`endif
        .border_number(border_number),              // input
        .text_number(text_number),                      // input
        .back_number(back_number),                      // input
        .dhrcolor_number(dhrcolor_number),          // input    
        .nibble(low_color_nibble),                      // input
        .dlr_nibble(dlr_nibble),                        // input
        .border_color12(border_color12),                // output
        .text_color12(text_color12),                    // output
        .back_color12(back_color12),                    // output
        .dhr_color12(dhr_color12),                      // output
        .lowres_color_val12(lowres_color_val12),    // output
        .dlr_color12(dlr_color12)                       // output
            );
//////////////////////////////////////////////////////////////////////////////////
// allow_shift7 determines when it's OK to shift a bit (in non-SHR modes)
    wire allow_shift7 = (mode_80STORE & mode_80COL | DHR) ? 1'b1: h_count[0]; 
//////////////////////////////////////////////////////////////////////////////////  
// creates the pixel clock.
    always @(posedge clk50) clk1 <= ~clk1;          // divides clk50 by 2 - pixelclock
//////////////////////////////////////////////////////////////////////////////////  
// increments the counters used for positioning the pixel output, blanking 
// and sync pulses.
// increments counters used in calculating the pixel data and shifting.
// calculates the ROM address for TEXT bits
    reg line_dbl = 0;                   // toggles for line doubling    
   always @(posedge clk50) 
        begin
        if(clk1)                                            // pixel clock is high                      
            begin               
// SHR address calc & line doubler
            if(disp40_active)                               // only if display is active
                begin                                   
                if (h_count[1:0] == 2'b00)              // only on every 4th count  
                    begin
                    shr_addr <= shr_addr + 1;           // increment the address counter
                    if(shr_addr == 16'h9FFF) shr_addr <= 16'h2000;  // restart counter
                    shr_bytes <= shr_bytes + 1;     // increment byte in line counter                       
                    if (shr_bytes == 8'h9F)             // if end of line
                        begin
                        if (line_dbl == 0) shr_addr <= shr_addr - shr_bytes; // line doubler
                        line_dbl <= ~line_dbl;          // flip doubler flag
                        end
                    end
                end
// DHR counter
            if ((h_count < 9 + test42) || (count28 == 27) || (h_count > 628 )) count28 <= 0; // reset counter
            else count28 <= count28 + 1;                // increment DHR 0-27 counter

// reset counters
            if (h_count == 909)                             // end of horizontal count 0-899  
                begin                                           // 900 is count for 720 wide
                h_count <= 0;                               // start over
                count7 <= 0;                                // reset 0-6 counter
        //      count28 <= 0;
                shr_bytes <= 8'h00;
// SHR vertical position                    
                if ((v_count == 447) || (v_count < 48)) 
                    begin
                    shr_addr <= 16'h2000;
                    line_dbl <= 1'b0;
                    end
// line counter
                if(v_count == 524) v_count <= 0;        // last line 0-524 start over
                else v_count <= v_count + 1;            // not last line so start next line
                end                         
// horizontal counter
            else                                                // not end of horizontal count 0-799
                begin
                h_count <= h_count + 1;                 // increment the horizontal count 0-799
// counter for shifting
                if(allow_shift7)                            // every clk1 for 80c, every other clk1 for 40c
                    begin
                    if(count7 == 6) count7 <= 0;        // end of 0 to 6 count so reset 
                    else count7 <= count7 + 1;          // not end so increment 
                    end    
                end
// character ROM address
//  was -   if(count7 == 3'b0) rom_addr <= {ram_data[7:0], scanline};
// the new coding fixes a long standing bug in the flashing text                                                                                                    
            if(count7 == 3'b0) rom_addr <= {ram_data[7], (ram_data[6] & AltCharSet) | (ram_data[7] & ram_data[6]), ram_data[5:0], scanline};    // sent character address to cb1.v  
            end
        end 
//////////////////////////////////////////////////////////////////////////////////      
// Aux memory - determine when to read or write aux memory (bank1)
// the commented-out code in load_aux_mem helped DazzleDraw slightly but caused other problems
    reg load_aux_mem = 0;
    reg aux_byte = 0;

    always @(posedge clk50)
        begin 
        load_aux_mem <= (mode_80COL && (TEXT | ( MIXED & vc[4] & vc[2]) | DHR ) && aux_byte) /* | (mode_80STORE & PAGE2) | (WRITERAM_SEL)*/ | (SHR);    // read from aux
        auxmem <= ((mode_80STORE & PAGE2) | (WRITERAM_SEL) | (m2b0_sl & (user1_s == 0)));                   // write to aux  
        end     
//////////////////////////////////////////////////////////////////////////////////      
// calculate video RAM addresses
    wire X_PAGE2 = PAGE2 & ~mode_80STORE;           
    wire [5:0] high_vid_addr = (hgr) ? {1'b0, X_PAGE2, ~X_PAGE2, scanline} : {4'b0000, X_PAGE2, ~X_PAGE2};  // line               
    wire [3:0] addr_sum = {~hc[5],~hc[5],hc[4],hc[3]} +  {vc[4],vc[3],vc[4],vc[3]} + 1;     
    wire [15:0] apple_video_addr = (SHR) ? shr_addr : {high_vid_addr, vc[2:0], addr_sum ,hc[2:0]};
   reg [15:0]  apple_video_addr_reg = 16'h0000;
    
    assign ram_address = {load_aux_mem, apple_video_addr_reg};  // this is the one used for getting ram_data from buffer 
//////////////////////////////////////////////////////////////////////////////////  
// border, text & background colors

    always @(*) 
        begin
`ifdef bbug             
        border_number = ((BORDER_COLOR_NUM + bugborder) > 15) ? 0 : BORDER_COLOR_NUM + bugborder;
`else               
        border_number = BORDER_COLOR_NUM;  
`endif
        text_number   = TEXT_COLOR_NUM;     
        back_number   = BACK_COLOR_NUM;     
        end
//////////////////////////////////////////////////////////////////////////////////
// active display timing

    reg hor_del1 = 0;
    reg [5:0] hor_del0 = 0;

    always @(posedge clk50)
        begin
// Super High Res graphics
        if (SHR || LINEAR_SHR && ~TEXT & ~clk1) // SHR or Linear SHR 
                begin
// the following delays were taken out because they weren't working correctly
// this needs to be checked closer
/*
                hor_active_reg <= hor_active;       // register the active signal
                hor_del0[0] <= hor_active_reg;  // delay
                hor_del0[1] <= hor_del0[0];     // delay
                hor_del0[2] <= hor_del0[1];     // delay
                hor_del0[3] <= hor_del0[2];     // delay
                hor_del0[4] <= hor_del0[3];     // delay
        //      hor_del0[5] <= hor_del0[4];
        //      hor_del2 <= hor_active;
        //      hor_del2 <= hor_del0[0];            // delay
*/
                apple_video_addr_reg <= apple_video_addr; // register the address
                end 
        else 
// text & low res graphics
      if(clk1 && allow_shift7 && ~hgr && ~SHR)          // every clk1 for 80c, every other clk1 for 40c
            begin                       
            if (count7 == 0)                                        // only once/byte
                begin
                hor_del2 <= hor_active_reg;                 // delay
                hor_active_reg <= hor_active;                   // register the active signal
                apple_video_addr_reg <= apple_video_addr;   // register the video address           
                end
            end 
        else
// high res graphics        
      if(clk1 && hgr && ~SHR)               // for HGR
            begin                       
            if ((count7 == 0) && ~mode_80COL && h_count[0])         
                begin
                hor_active_reg <= hor_active;           // register the signal
                hor_del2 <= hor_active_reg;             // delay
                apple_video_addr_reg <= apple_video_addr;               
                end
            else
// double high res graphics & high res 80 columns
            if ((count7 == 0) && mode_80COL)
                begin
                hor_active_reg <= hor_active;                   // register the active signal
                if (DHR & hgr)
                    begin
                    hor_del1 <= hor_active_reg;
                    end
                else                
                    hor_del1 <= hor_active_reg;                     // delay
                apple_video_addr_reg <= apple_video_addr;   // register the video address
                end
            if (mode_80COL) hor_del2 <= hor_del1;           // more delay with 80 columns
            end 
   end              
//////////////////////////////////////////////////////////////////////////////////
// Super Hires graphics (SHR) data processing
// 640 mode, 320 mode and Fill (scanline interrupts not supported)
//
// scb_data is scanline control byte from BRAM via cb1.v
// scb_addr is scanline control byte address in BRAM sent to cb1.v to retrieve scb_data
// pal_data is 12 bit palette color from BRAM via cb1.v
// pal_addr is palette color address in BRAM sent to cb1.v to retrieve pal_data
// shr_color12 is the actual color for video output

    reg [7:0] shr_data = 0;                         // local register to hold data from SRAM
    wire shr_resmode = scb_data[7];                 // 1 if 640 wide resolution.  0 if 320 wide. 
// calculate when to use Fill 
    wire fill = (shr_resmode == 0) & (scb_data[5] == 1) & (((h_count[1] == 1'b0) && 
                    (shr_data[7:4] == 0)) || ((h_count[1] == 1'b1) && (shr_data[3:0] == 0)));
    wire [3:0] palette_addr = scb_data[3:0];    // use lower 4 bits of scanline control byte for palette address
    wire [8:0] vfix = v_count - 48;             // was 47 but 48 fixes some of palette glitches 
    wire [7:0] a2line = vfix[8:1];              // vfix/2
`ifdef dither_640
    wire [11:0] dith;
    wire do_dither;
    wire dith640 = shr_resmode & CBVID[0];
//  wire [11:0] d0 = pal_data;
    reg [11:0] d0 = 12'h000;
    reg [11:0] d1 = 12'h000;
`endif
    always @(posedge clk50)
        begin
        if(SHR & clk1)                                  // when SHR is active and pixel clock is high
            begin   
            if(h_count[1:0] == 2'b11) shr_data <= ram_data[7:0];    // get data from SRAM via cb1.v every 4 pixels
`ifdef dither_640
            if(dith640)
                begin
                if(do_dither) shr_color12 <= dith;      // dither (dith is d1 mixed with d0)            
                else shr_color12 <= d1;                      // when not dithering
                end
            else
`endif  
            if (~fill) shr_color12 <= pal_data;         // if not Fill mode update the color from BRAM via cb1.v
            end                                                 // after shifting to sync with dith
        end
// calculate address of scanline control byte in BRAM
    assign scb_addr = a2line;                       
// calculate palette address in BRAM    
    assign pal_addr = (shr_resmode == 1) ? ((h_count[1:0] == 2'b00) ? {palette_addr[3:0], 2'b10, shr_data[7:6]}:
                            (h_count[1:0] == 2'b01) ? {palette_addr[3:0], 2'b11, shr_data[5:4]} :
                            (h_count[1:0] == 2'b10) ? {palette_addr[3:0], 2'b00, shr_data[3:2]} :
                            {palette_addr[3:0], 2'b01,shr_data[1:0]}):
                            ((h_count[1] == 1'b0) ? {palette_addr[3:0], shr_data[7:4]} : {palette_addr[3:0], shr_data[3:0]});
`ifdef dither_640
// dither - synthetic dither creating the 640 mode master colors
    always @(posedge clk1)
        begin
        d0 <= pal_data;     // puts the color from BRAM into a register (this will be the next pixel color)
        d1 <= d0;               // shift next color to the current pixel color
        end
    wire [4:0] r_dith = (d0[11:8] + d1[11:8]) >> 1;     // average the red value of the 2 adjacent pixels
    wire [4:0] g_dith = (d0[7:4] + d1[7:4]) >> 1;       // average the green value of the 2 adjacent pixels
    wire [4:0] b_dith = (d0[3:0] + d1[3:0]) >> 1;       // average the blue value of the 2 adjacent pixels
    assign do_dither =  (h_count[0]) ? (d1 != d0) : 1'bz;   // although this implies an internal tri-state xst fixes it 
    assign dith =   { r_dith[3:0], g_dith[3:0], b_dith[3:0] };
//  assign dith =   { ((d0[11:8] == d1[11:8]) ? d1[11:8] : 4'b1000), ((d0[7:4] == d1[7:4]) ? d1[7:4] : 4'b1000), ((d0[3:0] == d1[3:0]) ? d1[3:0] : 4'b1000) };

`endif
//////////////////////////////////////////////////////////////////////////////////
// Double High Resolution graphics (DHR)
//////////////////////////////////////////////////////////////////////////////////
// these registers hold the 7 bit color data from SRAM
    reg [6:0] aux_data0 = 0;
    reg [6:0] main_data1 = 0;
    reg [6:0] aux_data2 = 0;
    reg [6:0] main_data3 = 0;
// time to get data from SRAM 
// (note these times make no sense but work)    
    wire ard0 = (count28 > 20) & (count28 < 26); 
    wire mrd1 = (count28 == 28) | (count28 < 5);  
    wire ard2 = (count28 > 6) & (count28 < 12); 
    wire mrd3 = (count28 > 13) & (count28 < 19);

    always @(posedge clk1)
        begin
        if (DHR) 
            begin
            if (ard0) aux_data0 <= ram_data[6:0];   // data from aux memory for 1st byte
            if (mrd1) main_data1 <= ram_data[6:0]; // data from main memory for 2nd byte
            if (ard2) aux_data2 <= ram_data[6:0];   // data from aux memory for 3rd byte
            if (mrd3) main_data3 <= ram_data[6:0];  // data from main memory for 4th byte
            end
        end 

    wire [3:0] prev_color;
    wire [3:0] next_color;
    wire [3:0] c_number;                    // color number determined  
    wire bitshift;
    reg [7:0] bit_stream = 0;
    wire new_color;                         // used to find color change
    reg [3:0] barrel = 0;                   // used to shift bits before being used as new color
    reg[2:0] pcount = 0;                    
// manipulate the bits to create a bit stream of 4 bit color index numbers
    always @(posedge clk50)
        begin
        if (clk1)
            begin
            case (count28[4:0])         
            5'b00000: bit_stream <= {main_data1[0], aux_data0[6:0]};
            5'b00100: bit_stream <= { main_data1[4:0], aux_data0[6:4] };
            5'b01000: bit_stream <= { aux_data2[1:0], main_data1[6:1] } ;
            5'b01100: bit_stream <= { aux_data2[5:0], main_data1[6:5] };
            5'b10000: bit_stream <= { main_data3[2:0], aux_data2[6:2] };
            5'b10100: bit_stream <= { main_data3[6:0], aux_data2[6] };
            5'b11000: bit_stream <= { aux_data0[3:0], main_data3[6:3] };
            5'b11100: ;             //  should never happen
            default:  bit_stream <= {bit_stream[0], bit_stream[7:1]};   // rotate right so we use the next 4 bits
            endcase
            end
        end

    assign new_color = (count28[1] == 1'b0) || bit_stream[5:4] == bit_stream[1:0] ; // better *
    assign bitshift = (count28[1:0] != 2'b00);
    assign prev_color = {dhrcolor_number[0], dhrcolor_number[2], dhrcolor_number[2], dhrcolor_number[3]};
    assign next_color = bit_stream[3:0];
    
    always @(posedge clk1)
        begin
        if (DHR & ~SHR & (h_count < 629))
            begin
            if (new_color)
                begin
                dhrcolor_number <= {barrel[0], barrel[1], barrel[2], barrel[3]};
                end
            end
        end

    always @(*)
        begin
        case (count28[1:0])
        2'b00: barrel = bit_stream[3:0];
        2'b01: barrel = {bit_stream[2], bit_stream[1], bit_stream[0], bit_stream[3]};
        2'b10: barrel = {bit_stream[1], bit_stream[0], bit_stream[3], bit_stream[2]};
        2'b11: barrel = {bit_stream[0], bit_stream[3], bit_stream[2], bit_stream[1]};
        endcase
        end
//////////////////////////////////////////////////////////////////////////////////
// puts data from SRAM into shifter for hires & text.  LoRes data uses color table directly
    wire video_bit = shifter7[0];               // used by hires
   reg  high_bit;                               // used by hires

    always @(posedge clk50)
        begin       
// high res graphics
        if(clk1 && h_count[0] && hgr &&  ~DHR  && ~SHR ) 
            begin
            vid_del <= video_bit;               // used by hires routines
            if (count7 == 6)            
                begin
                shifter7 <= ram_data[6:0];      // put SRAM data into shifter
                high_bit <= ram_data[7];        // used by hires routines
                end
            else                                        // count7 does equal 6 so right shift
                begin
                shifter7 <= {1'b0, shifter7[7:1]};  // right shift 1 bit
                end
            end
// text and low res graphics
        else if(clk1 && allow_shift7 && ~hgr  && ~SHR)          
                begin
                    if(count7 == 0) 
                        begin
                        shifter7 <= (~ram_data[7]) ? char_data [7:0] : ~char_data[7:0]; // text data into shifter
                        low_color_nibble <= v_count[3] ? ram_data[7:4] : ram_data[3:0]; // lores index number   
                        flash  <= AltCharSet ? 0 : (ram_data[6] & ~ram_data[7]);                // enable/disable flash 
                        inverse <=  ~ram_data[7];                                                       // enable/disable inverse
                        end
                    else                                // not count7 == 0 so right shift                       
                        begin
                        shifter7 <= {1'b0, shifter7[7:1]};  // right shift 1 bit
                        end      
                end
        end // always @(posedge clk50)
//////////////////////////////////////////////////////////////////////////////////
// determine next vertical character line
    wire next_vc = (clk1 & allow_shift7 & (count7 == 0)) & (hc == 64) & ((aux_byte == 0) | ~mode_80COL); 
//////////////////////////////////////////////////////////////////////////////////
// color definitions used for 6 bit hires
`define WHITE  6'b111111
`define BLACK  6'b000000
`define GREEN  6'b001100    // was 6'b001000
`define PURPLE 6'b110010
`define RED    6'b110100
`define BLUE   6'b000011  // was 6'b001011
//////////////////////////////////////////////////////////////////////////////////
// high resolution graphics (HGR)
    wire  odd_dot = h_count[1];

    reg [11:0] last12 = 12'h000;
    reg [5:0] last6 = 6'b0;
    reg non_color = 0;
    always @ (posedge clk1)
        begin
        last12 <= color12;
        last6 <= color6;
        if (video_bit && vid_del) 
            begin
            color12 <=12'hFFF;  // 12 bit
            color6 <= 6'b111111; // 6 bit
            if (shift_bitsum != 1) non_color <= 1'b1;
            end
        else    
        if(video_bit && ~vid_del)
            begin
            if (non_color)
                begin 
                color12 <= last12;
                color6 <= last6;
                non_color <= 1'b0;
                end
            else    
            if(odd_dot) 
                begin  
                if(high_bit) 
                    begin
                    color12 <= 12'h22F;   
                    color6 <= `BLUE;
                    end
            else 
                begin
                color12 <= 12'hD2D;
                color6 <= `PURPLE;
                end
            end 
            else
                begin
                if(high_bit) 
                    begin
                    color12 <= 12'hF60; // color12 <= 12'hF98;
                    color6 <= `RED;
                    end
                else 
                    begin
                    color12 <= 12'h1D0;
                    color6 <= `GREEN;
                    end
                end
            end
            else 
            if (~video_bit && ~vid_del) 
                begin
                color12 <=12'h000;  // 12 bit 
                color6 <= 6'b000000; // 6 bit
                non_color <= 1'b1;
                end
            end 
////////////////////////////////////////////////////////////////////////////////// 
// horizontal character count (set & increment)
// aux_byte flag (set & toggled)
// non-SHR graphics delays
    always @(posedge clk50)
        begin
        if(clk1 && allow_shift7)    // every clk1 for 80c, every other clk1 for 40c
            begin
            if(~(hor_active && vert_active))    // video display is not active
                begin                                   // (we're in blanking)
                hc <= 7'b0011000;                       // start with 24 
                aux_byte <= (GR) ? 0 : 1;               // start with 0 for non-SHR graphics otherwise 1
                end     
            else                                            // video display is active
                begin                                                       
                if(count7 == 0)
                    begin
                    aux_byte <= ~aux_byte;          // flip aux_byte with each 7 bits
                    GR_r <= GR;                         // delay graphics 7 pixel clock cycles
                    GR_r2 <= GR_r;                      // delay graphics 14 pixel clock cycles
                    if((aux_byte == 1) || ~mode_80COL)  // always if not 80 col. othewise every other time
                        begin                                               
                        hc <= hc + 1;                   // increment counter 24 to 63
                        end
                    end  
                end 
            end 
        end
//////////////////////////////////////////////////////////////////////////////////
// vertical character count (set & incremented)
// flash count (incremented)
   always @(posedge clk50) 
        begin
        if (~vert_active) vc <= 5'b00000;   // vertical part of video display not active 
        else                                            // vertical part of video display active                                                    
            begin 
            if(next_vc)                              
                begin 
                if((v_count == 48) || (vc == 24))   // end of top border or start of bottom border  
                    begin
                    flash_count <= flash_count +1;  // increment flash count
                    vc <= 5'b00000;                         // reset vertical character (line) count
                    end   
                else  
                    if (v_count[3:0] == 4'b1111)  vc <= vc + 1;  // increment char line every 16 scan lines
                end 
            end 
        end 
////////////////////////////////////////////////////////////////////////////////// 
// decoding Apple soft switches 
    reg C3_ROM = 0;
    reg A2VBL = 1;
   wire io_access = (bus_addr[15:8] == 8'b11000000)/* & (user1_s == 0)*/;  // $C0xx area being accessed
    
    always @(negedge Q3_s)                          // negative of Apple Q3 clock
        begin 
        if(phase0_s == 0)   m2b0_sl <= m2b0_s;  // Apple phase0 clock low so latch m2b0 per IIgs technote #68                       
        if(ph1_s == 0)                                  // Apple phase1 clock low
            begin
//          io_read <= 0;
            if(reset_s == 1'b0)                         // Apple is RESET
                begin
//////////////////////////////////////////////////////////////////////////////////
// reset
// set registers used by vga_video.v to default values              
`ifdef tester
                test42 <= 8'b00000000;
`endif
                TEXT <= 1;
                PAGE2 <= 0;
                HIRES <= 0;
                MIXED <= 0;  
                AltCharSet <= 0;
                mode_80COL <= 0;
                mode_80STORE <= 0;
                DHR <= 0;
                SHR <= 0;
                WRITERAM_SEL <= 0;
                BORDER_COLOR_NUM <= 4'h0;  // text, background & border are black & white with reset pressed
                TEXT_COLOR_NUM <= 4'hF;
                BACK_COLOR_NUM <= 4'h0;     
                end 
//////////////////////////////////////////////////////////////////////////////////
// not reset
            else            
            if(io_access && (rw_s | ~rw_s))  
// read or write to $C0xx area
                begin   
                if (user1_s == 0)       // aka /M2SEL - 0 means in first 128KB of RAM
                    begin
// bank 0 or 1 access                   
                    case(bus_addr[7:0])                     // read or write of softswitch we use
`ifdef tester
                    8'h00:      KEY <=  bus_data[7:0];  // gets keypress code for read or write
                    8'h10:      KEY[7] <= 1'b0;         // strobe key press in CB
`endif
                    8'h0A:      C3_ROM <= 1;
                    8'h0B:      C3_ROM <= 0;
                    8'h19:      A2VBL <= ~bus_data[7];
                    8'h50:      TEXT <= 0;                  
                    8'h51:      TEXT <= 1;
                    8'h52:      MIXED <= 0;
                    8'h53:      MIXED <= 1;
                    8'h54:      PAGE2 <= 0;
                    8'h55:      PAGE2 <= 1;   
                    8'h56:      HIRES <= 0;
                    8'h57:      HIRES <= 1;               
                    8'h5E:        DHR <= 1;     // double res on 
                    8'h5F:        DHR <= 0;     // double res off
                    endcase 
                    if(rw_s == 1'b0)                // write
                        begin           
// write to softswitch
                        case(bus_addr[7:0])                 // write to softswitch we use                   
                        8'h00: mode_80STORE <= 0;           // $C000
                        8'h01: mode_80STORE <= 1;
                        8'h04: WRITERAM_SEL <= 0;
                        8'h05: WRITERAM_SEL <= 1;
                        8'h0C: mode_80COL <= 0;
                        8'h0D: mode_80COL <= 1;
                        8'h0E: AltCharSet <= 0;
                        8'h0F: AltCharSet <= 1;             // $C00F
                        8'h29:                                  // $C029  
                            begin   
                            SHR <= bus_data[7];             // state of super hires on/off              
                            LINEAR_SHR <= bus_data[6];      // state of linear SHR mode
                            monocolor <= bus_data[5];       // monocolor for DHR
                            end                     
                        8'h34:  BORDER_COLOR_NUM <= bus_data[3:0]; // $C034 lower 4 bits is color#
`ifdef tester
                        8'h42: test42 <= bus_data[7:0]; // input used for various tests
`endif
                        8'h22:  // $C022 high 4 bits = text color# low 4 bits = background color#
                            begin
                            TEXT_COLOR_NUM <= bus_data[7:4];    // high 4 bits = text color#
                            BACK_COLOR_NUM <= bus_data[3:0];    // low 4 bits = background color#
                            end
                        8'h68: 
                            begin
                            WRITERAM_SEL <= bus_data[4];        // Apple IIgs fast reg for write bank 0 or 1
                            end 
                        endcase
                        end 
                    end
                end  
            end 
        end
//////////////////////////////////////////////////////////////////////////////////
// sync apple clocks with main CB clock (from DCM)
   always @(posedge clk50) 
        begin
        reset_1 <= reset_n;     // this probably shouldn't be done
        reset_s <= reset_1;     // it's sync'd in cb1.v
        end
//////////////////////////////////////////////////////////////////////////////////      
// key press input - reads key press and value
`ifdef tester
    reg key_state = 0;          // used in the state machine

    always @(posedge Q3_s)
        begin
        case (key_state)
        0: begin
            if(KEY[7] == 1'b1)  // key pressed - this is the same as bit 7 of $C000
                begin
                case (KEY[6:0]) // same as bits 0 - 6 in $C000
                7'h08:  begin
                            test00 = test00 - 1;    // left arrow so decrement test00 reg.
                            key_state = 1'b1;       // set the state so we don't come back here on next clock
                            end
                7'h12:  begin
                            test00 = 0;             // ctrl - r  resets the test reg.
                            key_state = 1'b1;       // set the state so we don't come back here on next clock
                            end
                7'h15:  begin
                            test00 = test00 + 1; // right arrow so increment test00 reg.
                            key_state = 1'b1;      // set the state so we don't come back here on next clock
                            end
                7'h1D:  begin                       // ctrl - ]  to toggle SHR 640 dither
                            CBVID[0] <= ~CBVID[0];
                            key_state = 1'b1;
                            end
                endcase
                end
            end
        1: begin
            if(KEY[7] == 1'b0) key_state = 1'b0;    // key not pressed so just wait until it is
            end
        endcase
        end
`endif  
//////////////////////////////////////////////////////////////////////////////////      
// changes border color to indicate reg or wire value
// bugborder is added to the border_color_num

`ifdef bbug

   always @(posedge clk50) 
        begin
// bugboarder
// replace the 1'b0 with any wire or reg and it
// allows you to increment the border color # by it's value 
// 0 - normal border   1 - increment the color
//  bugborder <= 1'b0;          // any single bit reg or wire can be placed here        
                                            // use bugboarder <= bugin; if bringing in value
                                            // from cb1.v
    bugborder <= bugin; 
        end
`endif
endmodule
//////////////////////////////////////////////////////////////////////////////////
